# Copyright 2025 Stack AV Co.
# SPDX-License-Identifier: Apache-2.0

"""Triton point cloud voxeliation."""

from dataclasses import dataclass

import torch
import triton

from conch.kernels.vision.voxelization import generate_dense_voxels_triton_kernel, generate_voxels_triton_kernel


@dataclass
class VoxelizationParameter:
    """Parameters."""

    min_range: tuple[float, float, float]
    max_range: tuple[float, float, float]
    voxel_dim: tuple[float, float, float]
    grid_dim: tuple[int, int, int]
    max_num_points_per_voxel: int
    max_num_voxels: int

    def __init__(
        self,
        min_range: tuple[float, float, float],
        max_range: tuple[float, float, float],
        voxel_dim: tuple[float, float, float],
        max_num_points_per_voxel: int,
    ) -> None:
        """Init parameters."""
        self.min_range = min_range
        self.max_range = max_range
        self.voxel_dim = voxel_dim
        self.max_num_points_per_voxel = max_num_points_per_voxel
        self.grid_dim = self._compute_grid_dim()
        self.max_num_voxels = self.grid_dim[0] * self.grid_dim[1] * self.grid_dim[2]

    def _compute_grid_dim(self) -> tuple[int, int, int]:
        """Compute grid dimensions."""
        grid_x = round((self.max_range[0] - self.min_range[0]) / self.voxel_dim[0])
        grid_y = round((self.max_range[1] - self.min_range[1]) / self.voxel_dim[1])
        grid_z = round((self.max_range[2] - self.min_range[2]) / self.voxel_dim[2])
        return (grid_x, grid_y, grid_z)


def generate_voxels(
    points: torch.Tensor, param: VoxelizationParameter
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Generates voxels from input points, output voxels and points are randomly ordered due to use of atomics.

    Args:
        points: input points; expected dimensions (num_points, 4), last dimension should have fields of x,y,z,_.
        param: parameters.

    Returns:
        tuple of tensors:
            num_of_filled_voxels, shape [1, cpu]
            point_features, shape [num_filled_voxels, max_num_points_per_voxel, num_features_per_point], empty points are filled with 0.
            voxel_indices, shape [num_filled_voxels, 4], only first 3 fields are used for x,y,z indices.
            capped_num_points_per_voxel, shape [num_filled_voxels], note per voxel point counters are capped with max_num_points_per_voxel.
    """
    device = points.device
    num_points, num_features_per_point = points.shape
    assert num_features_per_point == 4
    # same as original nvidia cuda impl
    num_elements_per_voxel_index = 4

    # dense
    dense_num_points_per_voxel = torch.zeros((param.max_num_voxels), dtype=torch.int32, device=device)
    dense_point_features = torch.empty(
        (param.max_num_voxels, param.max_num_points_per_voxel, num_features_per_point), dtype=torch.float, device=device
    )

    # sparse/contiguous output
    num_filled_voxels = torch.zeros((1), dtype=torch.int32, device=device)
    num_points_per_voxel = torch.empty_like(dense_num_points_per_voxel)
    point_features = torch.empty_like(dense_point_features)
    voxel_indices = torch.empty((param.max_num_voxels, num_elements_per_voxel_index), dtype=torch.int32, device=device)

    # hardcoded for 256 threads per program save as cuda
    block_size = 256
    num_threads_per_warp = 32

    # first generate dense voxels
    grid = (triton.cdiv(num_points, block_size),)
    generate_dense_voxels_triton_kernel[grid](
        points,
        num_points,
        param.min_range[0],
        param.min_range[1],
        param.min_range[2],
        param.max_range[0],
        param.max_range[1],
        param.max_range[2],
        param.voxel_dim[0],
        param.voxel_dim[1],
        param.voxel_dim[2],
        param.grid_dim[0],
        param.grid_dim[1],
        param.grid_dim[2],
        param.max_num_points_per_voxel,
        dense_num_points_per_voxel,
        dense_point_features,
        cxpr_block_size=block_size,
        num_warps=block_size // num_threads_per_warp,
    )

    # compress into contiguous/sparse filled voxels
    grid = (triton.cdiv(param.max_num_voxels, block_size),)
    generate_voxels_triton_kernel[grid](
        dense_point_features,
        dense_num_points_per_voxel,
        param.grid_dim[0],
        param.grid_dim[1],
        param.max_num_points_per_voxel,
        param.max_num_voxels,
        num_filled_voxels,
        num_points_per_voxel,
        point_features,
        voxel_indices,
        cxpr_block_size=block_size,
        num_warps=block_size // num_threads_per_warp,
    )

    total_filled_voxels = num_filled_voxels.cpu()

    return total_filled_voxels, point_features, voxel_indices, num_points_per_voxel



# Copyright (c) 2025, Tri Dao.
# As of 2025-04-23, we require triton >= 3.0

from typing import Optional, Union

import torch

import triton
import triton.language as tl


@triton.jit
def rotary_kernel(
    OUT,  # Pointers to matrices
    X,
    COS,
    SIN,
    CU_SEQLENS,
    SEQLEN_OFFSETS,  # this could be int or a pointer
    # Matrix dimensions
    seqlen,
    nheads,
    seqlen_ro,
    # strides
    stride_out_batch,
    stride_out_seqlen,
    stride_out_nheads,
    stride_out_headdim,
    stride_x_batch,
    stride_x_seqlen,
    stride_x_nheads,
    stride_x_headdim,
    # Meta-parameters
    # We want ROTARY_DIM to be constexpr, otherwise the triton compiler doesn't know that
    # the mask is constant every 8 elements, and it will generate LDG.16 instead of LDG.128
    ROTARY_DIM: tl.constexpr,
    IS_SEQLEN_OFFSETS_TENSOR: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    INTERLEAVED: tl.constexpr,
    CONJUGATE: tl.constexpr,
    BLOCK_H: tl.constexpr,
    BLOCK_M: tl.constexpr,
):
    BLOCK_K: tl.constexpr = triton.next_power_of_2(ROTARY_DIM)
    ROTARY_DIM_HALF = ROTARY_DIM // 2
    pid_head = tl.program_id(axis=0)
    pid_m = tl.program_id(axis=1)
    pid_batch = tl.program_id(axis=2)

    if not IS_VARLEN:
        X = X + pid_batch * stride_x_batch
        OUT = OUT + pid_batch * stride_out_batch
    else:
        start_idx = tl.load(CU_SEQLENS + pid_batch)
        seqlen = tl.load(CU_SEQLENS + pid_batch + 1) - start_idx
        X = X + start_idx * stride_x_seqlen
        OUT = OUT + start_idx * stride_out_seqlen

    if pid_m * BLOCK_M >= seqlen:
        return

    rh = pid_head * BLOCK_H + tl.arange(0, BLOCK_H)
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    if not IS_SEQLEN_OFFSETS_TENSOR:
        rm_cs = rm + SEQLEN_OFFSETS
    else:
        rm_cs = rm + tl.load(SEQLEN_OFFSETS + pid_batch)

    rk_half = tl.arange(0, BLOCK_K // 2)
    COS = COS + (rm_cs[:, None] * ROTARY_DIM_HALF + rk_half[None, :])
    SIN = SIN + (rm_cs[:, None] * ROTARY_DIM_HALF + rk_half[None, :])
    mask_cs = (rm_cs[:, None] < seqlen_ro) & (rk_half[None, :] < ROTARY_DIM_HALF)
    cos = tl.load(COS, mask=mask_cs, other=1.0).to(tl.float32)
    sin = tl.load(SIN, mask=mask_cs, other=0.0).to(tl.float32)
    if CONJUGATE:
        sin = -sin

    if not INTERLEAVED:
        # Load the 1st and 2nd halves of X, do calculation, then store to 1st and 2nd halves of OUT
        X = X + (rh[:, None, None] * stride_x_nheads + rm[None, :, None] * stride_x_seqlen + rk_half[None, None, :] * stride_x_headdim)
        OUT = OUT + (rh[:, None, None] * stride_out_nheads + rm[None, :, None] * stride_out_seqlen + rk_half[None, None, :] * stride_out_headdim)
        mask = (rh[:, None, None] < nheads) & (rm[None, :, None] < seqlen) & (rk_half[None, None, :] < ROTARY_DIM_HALF)
        x0 = tl.load(X, mask=mask, other=0.0).to(tl.float32)
        x1 = tl.load(X + ROTARY_DIM_HALF * stride_x_headdim, mask=mask, other=0.0,).to(tl.float32)
        o0 = x0 * cos - x1 * sin
        o1 = x0 * sin + x1 * cos
        tl.store(OUT, o0, mask=mask)
        tl.store(OUT + ROTARY_DIM_HALF * stride_out_headdim, o1, mask=mask)
    else:
        rk = tl.arange(0, BLOCK_K)
        X = X + (rh[:, None, None] * stride_x_nheads + rm[None, :, None] * stride_x_seqlen + rk[None, None, :] * stride_x_headdim)
        OUT = OUT + (rh[:, None, None] * stride_out_nheads + rm[None, :, None] * stride_out_seqlen + rk[None, None, :] * stride_out_headdim)
        mask = (rh[:, None, None] < nheads) & (rm[None, :, None] < seqlen) & (rk[None, None, :] < ROTARY_DIM)
        x = tl.load(X, mask=mask, other=0.0).to(tl.float32)
        x0, x1 = tl.split(tl.reshape(x, [BLOCK_H, BLOCK_M, BLOCK_K // 2, 2]))
        o0 = x0 * cos - x1 * sin
        o1 = x0 * sin + x1 * cos
        o = tl.reshape(tl.join(o0, o1), [BLOCK_H, BLOCK_M, BLOCK_K])
        tl.store(OUT, o, mask=mask)


def apply_rotary(
    x: torch.Tensor,
    cos: torch.Tensor,
    sin: torch.Tensor,
    seqlen_offsets: Union[int, torch.Tensor] = 0,
    cu_seqlens: Optional[torch.Tensor] = None,
    max_seqlen: Optional[int] = None,
    interleaved=False,
    inplace=False,
    conjugate=False,
) -> torch.Tensor:
    """
    Arguments:
        x: (batch, seqlen, nheads, headdim) if cu_seqlens is None
            else (total_seqlen, nheads, headdim).
        cos: (seqlen_ro, rotary_dim / 2)
        sin: (seqlen_ro, rotary_dim / 2)
        seqlen_offsets: integer or integer tensor of size (batch,)
        cu_seqlens: (batch + 1,) or None
        max_seqlen: int
    Returns:
        y: (batch, seqlen, nheads, headdim)
    """
    is_varlen = cu_seqlens is not None
    if not is_varlen:
        batch, seqlen, nheads, headdim = x.shape
    else:
        assert max_seqlen is not None, "If cu_seqlens is passed in, then max_seqlen must be passed"
        total_seqlen, nheads, headdim = x.shape
        batch_p_1 = cu_seqlens.shape[0]
        batch = batch_p_1 - 1
        seqlen = max_seqlen
    seqlen_ro, rotary_dim = cos.shape
    assert sin.shape == cos.shape
    rotary_dim *= 2
    assert rotary_dim <= headdim, "rotary_dim must be <= headdim"
    assert headdim <= 256, "Only support headdim <= 256"
    assert seqlen_ro >= seqlen, "seqlen_ro must be >= seqlen"

    cos, sin = cos.contiguous(), sin.contiguous()
    if isinstance(seqlen_offsets, torch.Tensor):
        assert seqlen_offsets.shape == (batch,)
        assert seqlen_offsets.dtype in [torch.int32, torch.int64]
        seqlen_offsets = seqlen_offsets.contiguous()
    else:
        assert seqlen_offsets + seqlen <= seqlen_ro

    output = torch.empty_like(x) if not inplace else x
    if rotary_dim < headdim and not inplace:
        output[..., rotary_dim:].copy_(x[..., rotary_dim:])

    grid = lambda META: (triton.cdiv(nheads, META["BLOCK_H"]), triton.cdiv(seqlen, META["BLOCK_M"]), batch)  # noqa
    BLOCK_M = 8 if rotary_dim <= 128 else 4

    # Need this, otherwise Triton tries to launch from cuda:0 and we get
    # ValueError: Pointer argument (at 0) cannot be accessed from Triton (cpu tensor?)
    with torch.cuda.device(x.device.index):
        torch.library.wrap_triton(rotary_kernel)[grid](
            output,  # data ptrs
            x,
            cos,
            sin,
            cu_seqlens,
            seqlen_offsets,
            seqlen,  # shapes
            nheads,
            seqlen_ro,
            output.stride(0) if not is_varlen else 0,  # batch_strides if not varlen else 0
            output.stride(-3),  # seqlen_stride or total_seqlen_stride
            output.stride(-2),  # nheads_stride
            output.stride(-1),  # headdim_stride
            x.stride(0) if not is_varlen else 0,  # batch_strides if not varlen else 0
            x.stride(-3),  # seqlen stride or total_seqlen_stride
            x.stride(-2),  # nheads stride
            x.stride(-1),  # headdim stride
            rotary_dim,
            isinstance(seqlen_offsets, torch.Tensor),
            is_varlen,
            interleaved,
            conjugate,
            BLOCK_M=BLOCK_M,
            BLOCK_H=2,
        )
    return output



import logging

import torch
import triton
import triton.language as tl

from flag_gems.ops.max import max_kernel_1, max_kernel_2
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import dim_compress, libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
# torch.any: Tests if any elements in input evaluate to True. If the dtype of input
#            is not BOOL, then test if any elements in input evaluate to non-zero value
# In triton function, test if any elements in input evaluate to non-zero value is ok.

cluster_num = 12
core_num = 64
thread_num = core_num * cluster_num
buf_len_per_core = 2048
vector_size = 16


def get_block(n: int) -> int:
    if n < cluster_num:
        res = cluster_num
    else:
        res = cluster_num * triton.cdiv(n, cluster_num)
    return res


def heur_m_block_size(args):
    return triton.next_power_of_2(min(triton.cdiv(args["M"], cluster_num), core_num))


def heur_n_block_size(args):
    return triton.next_power_of_2(min(args["N"], triton.cdiv(buf_len_per_core, 4)))


@triton.jit
def reduce_any(a, b):
    return a or b


@libentry()
# @triton.autotune(configs=runtime.get_tuned_config("any"), key=["M", "N"])
@triton.heuristics(
    values={
        "BLOCK_M": heur_m_block_size,
        "BLOCK_N": heur_n_block_size,
    },
)
@triton.jit
def any_kernel_dim(
    inp,
    out,
    M,
    N,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # Map the program id to the row of inp it should compute.
    pid = tle.program_id(0)
    rows = pid * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    inp = inp + rows * N
    out = out + rows
    row_mask = rows < M

    _any = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.int1)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask

        a = tl.load(inp + cols, mask, other=0.0)
        _any = _any or (a != 0)
    any = tl.reduce(_any, axis=1, combine_fn=reduce_any)
    tl.store(out, any[:, None], row_mask)


@libentry()
@triton.heuristics(
    values={
        "BLOCK_M": heur_m_block_size,
        "BLOCK_N": heur_n_block_size,
    },
)
@triton.jit
def max_kernel_dim(
    in_ptr,
    out_ptr,
    M,
    N,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    xoffset = tl.program_id(0) * BLOCK_M
    xindex = xoffset + tl.arange(0, BLOCK_M)[:, None]
    xmask = xindex < M
    rbase = tl.arange(0, BLOCK_N)[None, :]
    _max = tl.full([BLOCK_M, BLOCK_N], float("-inf"), tl.float32)
    for roffset in range(0, N, BLOCK_N):
        rindex = roffset + rbase
        rmask = rindex < N
        r1 = rindex
        inp = tl.load(
            in_ptr + (r1 + (N * xindex)), rmask & xmask, other=float("-inf")
        ).to(tl.float32)
        inpb = tl.broadcast_to(inp, [BLOCK_M, BLOCK_N])
        _max = tl.maximum(_max, inpb)
    tmp2 = tl.max(_max, axis=1, return_indices=False)[:, None]
    tl.store(out_ptr + xindex, tmp2, xmask)


@libentry()
@triton.jit
def any_kernel_1(
    inp,
    mid,
    n_elements,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < n_elements
    inp_val = tl.load(inp_ptrs, mask=mask, other=0.0)
    any_val = tl.reduce(inp_val != 0, axis=0, combine_fn=reduce_any)
    mid_ptr = mid + pid
    tl.store(mid_ptr, any_val)


@libentry()
@triton.jit
def any_kernel_2(mid, out, MID_SIZE, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid + offset
    mask = offset < MID_SIZE
    mid_val = tl.load(mid_ptrs, mask=mask, other=0).to(tl.int1)
    any_val = tl.reduce(mid_val, axis=0, combine_fn=reduce_any)
    tl.store(out, any_val)


def any(inp):
    logger.debug("GEMS ANY")
    n_elements = inp.numel()
    block_size = max(
        triton.cdiv(get_block(n_elements), cluster_num),
        triton.cdiv(buf_len_per_core * core_num, 4),
    )

    mid_size = triton.cdiv(n_elements, block_size)
    block_mid = triton.next_power_of_2(mid_size)

    if n_elements >= vector_size * thread_num:
        inp_uint8 = inp.view(torch.uint8)

        mid = torch.empty((mid_size,), dtype=torch.uint8, device=inp.device)
        out = torch.empty([], dtype=torch.uint8, device=inp.device)

        with torch_device_fn.device(inp.device):
            max_kernel_1[(mid_size, 1)](
                inp_uint8, mid, n_elements, block_size, buffer_size_limit=2048
            )
            if mid_size == 1:
                return mid.view(torch.bool).reshape([])

            max_kernel_2[(1, 1)](mid, out, mid_size, block_mid, buffer_size_limit=2048)
        out = out.view(torch.bool)
    else:
        mid = torch.empty((mid_size,), dtype=torch.bool, device=inp.device)
        out = torch.empty([], dtype=torch.bool, device=inp.device)

        with torch_device_fn.device(inp.device):
            any_kernel_1[(mid_size, 1)](
                inp, mid, n_elements, block_size, buffer_size_limit=2048
            )
            if mid_size == 1:
                return mid.reshape([])
            any_kernel_2[(1, 1)](mid, out, mid_size, block_mid, buffer_size_limit=2048)

    return out


def any_dim(inp, dim=None, keepdim=False):
    logger.debug("GEMS ANY DIM")
    shape = list(inp.shape)
    if dim is None:
        out = any(inp)
        if keepdim:
            out = torch.reshape(out, [1] * inp.ndim)
    else:
        assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
        dim = dim % inp.ndim
        inp = dim_compress(inp, dim)
        N = shape[dim]
        shape[dim] = 1
        M = inp.numel() // N

        if N >= vector_size * vector_size:
            # according to api, op == any, use max to calculate
            inpf = inp.to(torch.float)
            outf = torch.empty(shape, dtype=torch.float, device=inp.device)

            grid = lambda meta: (triton.cdiv(M, meta["BLOCK_M"]),)
            with torch_device_fn.device(inp.device):
                max_kernel_dim[grid](inpf, outf, M, N, buffer_size_limit=2048)
            out = outf.to(torch.bool)
        else:
            out = torch.empty(shape, dtype=torch.bool, device=inp.device)
            grid = lambda meta: (triton.cdiv(M, meta["BLOCK_M"]),)
            with torch_device_fn.device(inp.device):
                any_kernel_dim[grid](inp, out, M, N, buffer_size_limit=2048)

        if not keepdim:
            out = out.squeeze(dim=dim)
    return out


def any_dims(inp, dim=None, keepdim=False):
    logger.debug("GEMS ANY DIMS")

    if dim is None or isinstance(dim, int):
        return any_dim(inp, dim=dim, keepdim=keepdim)
    assert ((i >= -inp.ndim and i < inp.ndim) for i in dim), "Invalid dim"

    shape = list(inp.shape)
    dim = [d % inp.ndim for d in dim]
    inp = dim_compress(inp, dim)
    N = 1
    for i in dim:
        N *= shape[i]
        shape[i] = 1
    M = inp.numel() // N

    if N >= vector_size * core_num:
        # according to api, op == any, use max to calculate
        inpf = inp.to(torch.float)
        outf = torch.empty(shape, dtype=torch.float, device=inp.device)

        grid = lambda meta: (triton.cdiv(M, meta["BLOCK_M"]),)
        with torch_device_fn.device(inp.device):
            max_kernel_dim[grid](inpf, outf, M, N, buffer_size_limit=2048)
        out = outf.to(torch.bool)
    else:
        out = torch.empty(shape, dtype=torch.bool, device=inp.device)
        grid = lambda meta: (triton.cdiv(M, meta["BLOCK_M"]),)
        with torch_device_fn.device(inp.device):
            any_kernel_dim[grid](inp, out, M, N, buffer_size_limit=2048)

    if not keepdim:
        out = out.squeeze(dim=dim)
    return out





import torch
from einops import repeat


def naive_recurrent_gsa(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    s: torch.Tensor,
    g: torch.Tensor | None = None,
    scale: int | None = None,
    initial_state: torch.Tensor | None = None,
    output_final_state: bool | None = False,
) -> torch.Tensor:
    dtype = q.dtype
    q, k, v, s, g = map(lambda x: x.transpose(1, 2).contiguous().float(), (q, k, v, s, g))

    NG = q.shape[1]//k.shape[1]
    # [batch_size, n_heads, seq_len, n_slots]
    if g is None:
        z = s.float().logcumsumexp(2)
        g = torch.cat((z[:, :, :1], z[:, :, :-1]), 2) - z
        s = torch.exp(s - z)
    k, v, s, g = map(lambda x: repeat(x, 'b h t d -> b (h g) t d', g=NG), (k, v, s, g))
    if initial_state is not None:
        initial_state = tuple(map(lambda x: repeat(x, 'b h k v -> b (h g) k v', g=NG), initial_state))

    B, H, T, K, V, M = *q.shape, v.shape[-1], s.shape[-1]

    hk = torch.zeros(B, H, K, M, dtype=torch.float, device=q.device)
    ok = torch.zeros_like(s)

    if scale is None:
        scale = q.shape[-1] ** -0.5

    final_state = None
    if initial_state is not None:
        hk += initial_state[0]

    for i in range(T):
        q_i = q[:, :, i] * scale
        k_i = k[:, :, i]
        v_i = s[:, :, i]
        g_i = g[:, :, i].exp()
        hk = hk * g_i[..., None, :] + k_i[..., None] * v_i[..., None, :]
        ok[:, :, i] = (q_i[..., None] * hk).sum(-2)

    qv = ok.softmax(-1)
    hv = torch.zeros(B, H, M, V, dtype=torch.float, device=q.device)
    ov = torch.zeros_like(v)
    if initial_state is not None:
        hv += initial_state[1]

    for i in range(T):
        q_i = qv[:, :, i]
        k_i = s[:, :, i]
        v_i = v[:, :, i]
        g_i = g[:, :, i].exp()
        hv = hv * g_i[..., :, None] + k_i[..., None] * v_i[..., None, :]
        ov[:, :, i] = (q_i[..., None] * hv).sum(-2)

    if output_final_state:
        final_state = (hk.view(B, -1, NG, K, M)[:, :, 0], hv.view(B, -1, NG, M, V)[:, :, 0])
    ov = ov.transpose(1, 2).contiguous()
    return ov.to(dtype), final_state



"""
Utilities for attorch kernels and layers.
"""


from typing import List, Optional

import torch
import triton


def allow_tf32() -> bool:
    """
    Returns whether the current GPU architecture supports TF32.
    """
    return torch.cuda.get_device_capability()[0] >= 8


def get_n_stages(n_stages: int = 2) -> int:
    """
    Receives number of stages for software pipelining and returns it as-is
    if the GPU architecture is Ampere or newer and 2 otherwise.
    """
    return 2 if torch.cuda.get_device_capability()[0] < 8 else n_stages


def get_output_dtype(
    input_dtype: torch.dtype = torch.float32,
    autocast: Optional[str] = None,
    ) -> torch.dtype:
    """
    Returns the appropriate output dtype for automatic mixed precision
    given the input dtype and the operation's autocast behaviour.

    Args:
        input_dtype: Input dtype.
        autocast: The relevent operation's autocast behaviour.
            None signifies the input dtype should flow through,
            'fp16' signifies autocasting to FP16 when AMP is enabled,
            and 'fp32' signifies autocasting to FP32 when AMP is enabled.
    """
    dtype = torch.get_autocast_dtype('cuda')
    assert dtype, \
        f'Only autocast to float16 is supported, received {dtype}'

    if torch.is_autocast_enabled():
        if autocast is None:
            return input_dtype

        elif autocast == 'fp16':
            return torch.float16

        elif autocast == 'fp32':
            return torch.float32

        else:
            raise RuntimeError(f'Autocast type {autocast} is invalid. '
                               'Options are None, fp16, and fp32')

    else:
        return input_dtype


def element_wise_kernel_configs(
    block_name: str = 'BLOCK_SIZE',
    ) -> List[triton.Config]:
    """
    Returns kernel configurations for element-wise operations.

    Args:
        block_name: Name of block argument rows are distributed over.
    """
    return [triton.Config({block_name: 64}, num_warps=2),
            triton.Config({block_name: 128}, num_warps=2),
            triton.Config({block_name: 256}, num_warps=4),
            triton.Config({block_name: 512}, num_warps=4),
            triton.Config({block_name: 1024}, num_warps=4)]


def warps_kernel_configs() -> List[triton.Config]:
    """
    Returns kernel configurations with all possible number of warps.
    """
    return [triton.Config({}, num_warps=2**i) for i in range(6)]



# Copyright (c) 2024, Tri Dao, Albert Gu.

"""We want triton==2.1.0 or 2.2.0 for this
"""

import math
import torch
import torch.nn.functional as F

import triton
import triton.language as tl

from einops import rearrange, repeat

from mamba_ssm.utils.determinism import autotune_configs


def init_to_zero(names):
    return lambda nargs: [nargs[name].zero_() for name in names if nargs[name] is not None]


@triton.autotune(
    configs=autotune_configs([
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 64}, num_stages=3, num_warps=8),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 32}, num_stages=5, num_warps=2),
        triton.Config({'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32}, num_stages=5, num_warps=2),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32}, num_stages=4, num_warps=2),
    ]),
    key=['chunk_size', 'K', 'IS_CAUSAL'],
)
@triton.jit
def _bmm_chunk_fwd_kernel(
    # Pointers to matrices
    a_ptr, b_ptr, out_ptr, seq_idx_ptr,
    # Matrix dimensions
    seqlen, chunk_size, K, ngroups,
    stride_a_batch, stride_a_seqlen, stride_a_head, stride_ak,
    stride_b_batch, stride_b_seqlen, stride_b_head, stride_bk,
    stride_out_batch, stride_out_chunk, stride_out_head, stride_outm, stride_outn,
    stride_seq_idx_batch, stride_seq_idx_seqlen,
    # Meta-parameters
    IS_CAUSAL: tl.constexpr,
    dot_dtype: tl.constexpr,
    HAS_SEQ_IDX: tl.constexpr,
    BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,
):
    pid_b = tl.program_id(axis=1)
    pid_ch = tl.program_id(axis=2)
    pid_c = pid_ch // ngroups
    pid_h = pid_ch - pid_c * ngroups
    num_pid_n = tl.cdiv(chunk_size, BLOCK_SIZE_N)
    pid_m = tl.program_id(axis=0) // num_pid_n
    pid_n = tl.program_id(axis=0) % num_pid_n
    if IS_CAUSAL:
        if pid_n * BLOCK_SIZE_N >= (pid_m + 1) * BLOCK_SIZE_M:
            return
    a_ptr += pid_b * stride_a_batch + pid_c * chunk_size * stride_a_seqlen + pid_h * stride_a_head
    b_ptr += pid_b * stride_b_batch + pid_c * chunk_size * stride_b_seqlen + pid_h * stride_b_head
    if HAS_SEQ_IDX:
        seq_idx_ptr += pid_b * stride_seq_idx_batch + pid_c * chunk_size * stride_seq_idx_seqlen

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    a_ptrs = a_ptr + (offs_m[:, None] * stride_a_seqlen + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_n[None, :] * stride_b_seqlen)
    chunk_size_limit = min(chunk_size, seqlen - pid_c * chunk_size)

    acc = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        a = tl.load(a_ptrs, mask=(offs_m[:, None] < chunk_size_limit) & (offs_k[None, :] < K - k * BLOCK_SIZE_K), other=0.0).to(dot_dtype)
        b = tl.load(b_ptrs, mask=(offs_k[:, None] < K - k * BLOCK_SIZE_K) & (offs_n[None, :] < chunk_size_limit), other=0.0).to(dot_dtype)
        acc += tl.dot(a, b)
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * stride_bk

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    if HAS_SEQ_IDX:
        chunk_size_limit = min(chunk_size, seqlen - pid_c * chunk_size)
        seq_idx_m = tl.load(seq_idx_ptr + offs_m * stride_seq_idx_seqlen, mask=offs_m < chunk_size_limit, other=-1)
        seq_idx_n = tl.load(seq_idx_ptr + offs_n * stride_seq_idx_seqlen, mask=offs_n < chunk_size_limit, other=-2)
        acc = tl.where(seq_idx_m[:, None] == seq_idx_n[None, :], acc, 0.0)
    out = acc.to(out_ptr.dtype.element_ty)

    out_ptr += pid_b * stride_out_batch + pid_c * stride_out_chunk + pid_h * stride_out_head
    out_ptrs = out_ptr + (stride_outm * offs_m[:, None] + offs_n[None, :] * stride_outn)
    tl.store(out_ptrs, out, mask=(offs_m[:, None] < chunk_size) & (offs_n[None, :] < chunk_size))


@triton.autotune(
    configs=autotune_configs([
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_CS': 64}, num_stages=3, num_warps=8),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_CS': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_CS': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_CS': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_CS': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_CS': 32}, num_stages=4, num_warps=4),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_CS': 32}, num_stages=5, num_warps=2),
        triton.Config({'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_CS': 32}, num_stages=5, num_warps=2),
        triton.Config({'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_CS': 32}, num_stages=4, num_warps=2),
    ]),
    key=['chunk_size', 'K'],
)
@triton.jit
def _bmm_chunk_bwd_kernel(
    # Pointers to matrices
    a_ptr, dout_ptr, db_ptr, res_ptr,
    # Matrix dimensions
    seqlen, chunk_size, K, ngroups,
    stride_a_batch, stride_a_seqlen, stride_a_head, stride_ak,
    stride_dout_batch, stride_dout_chunk, stride_dout_head, stride_dout_csize_m, stride_dout_csize_n,
    stride_db_batch, stride_db_seqlen, stride_db_head, stride_db_k,
    stride_res_batch, stride_res_seqlen, stride_res_head, stride_res_k,
    # Meta-parameters
    dot_dtype: tl.constexpr,
    HAS_RESIDUAL: tl.constexpr,
    BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_CS: tl.constexpr,
):
    pid_b = tl.program_id(axis=1)
    pid_ch = tl.program_id(axis=2)
    pid_c = pid_ch // ngroups
    pid_h = pid_ch - pid_c * ngroups
    num_pid_n = tl.cdiv(K, BLOCK_SIZE_N)
    pid_m = tl.program_id(axis=0) // num_pid_n
    pid_n = tl.program_id(axis=0) % num_pid_n

    a_ptr += pid_b * stride_a_batch + pid_c * chunk_size * stride_a_seqlen + pid_h * stride_a_head
    dout_ptr += pid_b * stride_dout_batch + pid_c * stride_dout_chunk + pid_h * stride_dout_head

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_cs = tl.arange(0, BLOCK_SIZE_CS)
    dout_ptrs = dout_ptr + (offs_m[:, None] * stride_dout_csize_n + offs_cs[None, :] * stride_dout_csize_m)
    a_ptrs = a_ptr + (offs_cs[:, None] * stride_a_seqlen + offs_n[None, :] * stride_ak)
    chunk_size_limit = min(chunk_size, seqlen - pid_c * chunk_size)

    acc = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for cs in range(0, tl.cdiv(chunk_size_limit, BLOCK_SIZE_CS)):
        dout = tl.load(dout_ptrs, mask=(offs_m[:, None] < chunk_size) & (offs_cs[None, :] < chunk_size_limit - cs * BLOCK_SIZE_CS), other=0.0).to(dot_dtype)
        a = tl.load(a_ptrs, mask=(offs_cs[:, None] < chunk_size_limit - cs * BLOCK_SIZE_CS) & (offs_n[None, :] < K), other=0.0).to(dot_dtype)
        acc += tl.dot(dout, a)
        dout_ptrs += BLOCK_SIZE_CS * stride_dout_csize_m
        a_ptrs += BLOCK_SIZE_CS * stride_a_seqlen

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    if HAS_RESIDUAL:
        res_ptr += pid_b * stride_res_batch + pid_c * chunk_size * stride_res_seqlen + pid_h * stride_res_head
        res_ptrs = res_ptr + (offs_m[:, None] * stride_res_seqlen + offs_n[None, :] * stride_res_k)
        res = tl.load(res_ptrs, mask=(offs_m[:, None] < chunk_size_limit) & (offs_n[None, :] < K)).to(tl.float32)
        acc += res
    db = acc.to(db_ptr.dtype.element_ty)

    db_ptr += pid_b * stride_db_batch + pid_c * chunk_size * stride_db_seqlen + pid_h * stride_db_head
    db_ptrs = db_ptr + (offs_m[:, None] * stride_db_seqlen + offs_n[None, :] * stride_db_k)
    tl.store(db_ptrs, db, mask=(offs_m[:, None] < chunk_size_limit) & (offs_n[None, :] < K))


def _bmm_chunk_fwd(a, b, chunk_size, seq_idx=None, causal=False, output_dtype=None):
    """
    Argument:
        a: (batch, seqlen, k) or (batch, seqlen, ngroups, k)
        b: (batch, seqlen, k) or (batch, seqlen, ngroups, k)
        seq_idx: (batch, seqlen) or None. out[i, j] for seq_idx[i] != seq_idx[j] will be zeroed out.
        causal: if True, then out[i, j] for i > j will be arbitrary, only out[i, j] for i <= j are
            guaranteed to be correct.
    Return:
        out: (batch, nchunks, chunk_size, chunk_size) or (batch, nchunks, ngroups, chunk_size, chunk_size)
    """
    # Check constraints.
    has_groups = a.dim() == 4
    if not has_groups:
        batch, seqlen, k = a.shape
    else:
        batch, seqlen, ngroups, k = a.shape
    assert b.shape == a.shape
    if seq_idx is not None:
        assert seq_idx.shape == (batch, seqlen)
    if a.stride(-1) != 1 and a.stride(1) != 1:
        a = a.contiguous()
    if b.stride(-1) != 1 and b.stride(1) != 1:
        b = b.contiguous()
    nchunks = math.ceil(seqlen / chunk_size)
    # Allocates output.
    out_dtype = a.dtype if output_dtype is None else output_dtype
    out = torch.empty((batch, nchunks, chunk_size, chunk_size) if not has_groups else (batch, nchunks, ngroups, chunk_size, chunk_size),
                      device=a.device, dtype=out_dtype)
    dot_dtype = (tl.bfloat16 if a.dtype == torch.bfloat16 or b.dtype == torch.bfloat16 else
                 (tl.float16 if a.dtype == torch.float16 or b.dtype == torch.float16 else tl.float32))
    grid = lambda META: (triton.cdiv(chunk_size, META['BLOCK_SIZE_M']) * triton.cdiv(chunk_size, META['BLOCK_SIZE_N']),
                    batch, nchunks if not has_groups else nchunks * ngroups)
    with torch.cuda.device(a.device.index):
        _bmm_chunk_fwd_kernel[grid](
            a, b, out, seq_idx,
            seqlen, chunk_size, k, ngroups if has_groups else 1,
            a.stride(0), a.stride(1), 0 if not has_groups else a.stride(2), a.stride(-1),
            b.stride(0), b.stride(1), 0 if not has_groups else b.stride(2), b.stride(-1),
            out.stride(0), out.stride(1), 0 if not has_groups else out.stride(2), out.stride(-2), out.stride(-1),
            *((seq_idx.stride(0), seq_idx.stride(1)) if seq_idx is not None else (0, 0)),
            causal,
            dot_dtype,
            HAS_SEQ_IDX=seq_idx is not None,
        )
    return out


def _bmm_chunk_bwd(a, dout, residual=None, out=None):
    """
    Argument:
        a: (batch, seqlen, k) or (batch, seqlen, ngroups, k)
        dout: (batch, nchunks, chunk_size, chunk_size) or (batch, nchunks, ngroups, chunk_size, chunk_size)
        residual: (batch, seqlen, k) or (batch, seqlen, ngroups, k)
    Return:
        out: (batch, seqlen, k) or (batch, seqlen, ngroups, k)

    If there was seq_idx in the fwd pass, then dout[i, j] for seq_idx[i] != seq_idx[j] should already be
    zeroed out before calling this function.
    """
    # Check constraints.
    has_groups = a.dim() == 4
    if not has_groups:
        batch, seqlen, k = a.shape
    else:
        batch, seqlen, ngroups, k = a.shape
    nchunks, chunk_size = dout.shape[1], dout.shape[-1]
    if a.stride(-1) != 1 and a.stride(-2) != 1:
        a = a.contiguous()
    if dout.stride(-1) != 1 and dout.stride(-2) != 1:
        dout = dout.contiguous()
    if residual is not None:
        assert residual.shape == (batch, seqlen, k) if not has_groups else (batch, seqlen, ngroups, k)
        if residual.stride(-1) != 1 and residual.stride(1) != 1:
            residual = residual.contiguous()
    # Allocates output.
    if out is not None:
        assert out.shape == a.shape
        assert out.stride(-1) == 1 or out.stride(1) == 1
    else:
        out = torch.empty_like(a)
    dot_dtype = (tl.bfloat16 if a.dtype == torch.bfloat16 or dout.dtype == torch.bfloat16 else
                 (tl.float16 if a.dtype == torch.float16 or dout.dtype == torch.float16 else tl.float32))
    grid = lambda META: (triton.cdiv(chunk_size, META['BLOCK_SIZE_M']) * triton.cdiv(k, META['BLOCK_SIZE_N']), batch,
                    nchunks if not has_groups else nchunks * ngroups)
    residual_strides = ((residual.stride(0), residual.stride(1), 0 if not has_groups else residual.stride(2),
                         residual.stride(-1))
                        if residual is not None else (0, 0, 0, 0))
    with torch.cuda.device(a.device.index):
        _bmm_chunk_bwd_kernel[grid](
            a, dout, out, residual,
            seqlen, chunk_size, k, ngroups if has_groups else 1,
            a.stride(0), a.stride(1), 0 if not has_groups else a.stride(2), a.stride(-1),
            dout.stride(0), dout.stride(1), 0 if not has_groups else dout.stride(2), dout.stride(-2), dout.stride(-1),
            out.stride(0), out.stride(1), 0 if not has_groups else out.stride(2), out.stride(-1),
            residual_strides[0], residual_strides[1], residual_strides[2], residual_strides[3],
            dot_dtype,
            HAS_RESIDUAL=residual is not None,
        )
    return out



import logging
import os

import torch
import triton
import triton.language as tl

from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry, tl_extra_shim
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
rsqrt = tl_extra_shim.rsqrt


@libentry()
@triton.jit(do_not_specialize=["eps"])
def group_norm_kernel(
    X,
    Y,
    W,
    B,
    Mean,
    Rstd,
    group_size,
    C,
    HW,
    num_groups,
    eps,
    BLOCK_GROUP_SIZE: tl.constexpr,
    BLOCK_HW_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    group = pid % num_groups
    num_elements = group_size * HW
    group_offset = tl.arange(0, BLOCK_GROUP_SIZE)
    hw_offset = tl.arange(0, BLOCK_HW_SIZE)

    wb_offset = group * group_size + group_offset
    wb_mask = wb_offset < C

    xy_offset = pid * num_elements + group_offset[:, None] * HW + hw_offset[None, :]
    xy_mask = wb_offset[:, None] < C and hw_offset[None, :] < HW

    Mean_ptr = Mean + pid
    Rstd_ptr = Rstd + pid

    X_ptr = X + xy_offset
    Y_ptr = Y + xy_offset

    X_val = tl.load(X_ptr, mask=xy_mask, other=0.0).to(tl.float32)
    mean = tl.sum(X_val) / num_elements
    x = tl.where(xy_mask, X_val - mean, 0.0)

    var = tl.sum(x * x) / num_elements
    rstd = rsqrt(var + eps)
    x_hat = x * rstd

    if W is None:
        weight = 1
    else:
        weight = tl.load(W + wb_offset, mask=wb_mask, other=0.0)[:, None]
    if B is None:
        bias = 0
    else:
        bias = tl.load(B + wb_offset, mask=wb_mask, other=0.0)[:, None]
    Y_val = x_hat * weight + bias

    tl.store(Y_ptr, Y_val, mask=xy_mask)
    tl.store(Mean_ptr, mean)
    tl.store(Rstd_ptr, rstd)


@libentry()
@triton.jit
def group_norm_backward_kernel(
    grad_y,
    X,
    W,
    Mean,
    Rstd,
    num_groups,
    group_size,
    grad_x,
    C,
    HW: tl.constexpr,
    BLOCK_GROUP_SIZE: tl.constexpr,
    BLOCK_HW_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    group = pid % num_groups
    num_elements = group_size * HW

    group_offset = tl.arange(0, BLOCK_GROUP_SIZE)
    wb_offset = group * group_size + group_offset

    wb_mask = wb_offset < C

    rstd = tl.load(Rstd + pid).to(tl.float32)
    mean = tl.load(Mean + pid).to(tl.float32)
    if W is None:
        weight = 1
    else:
        weight = tl.load(W + wb_offset, mask=wb_mask, other=0.0).to(tl.float32)[:, None]

    dx_part2 = tl.zeros([BLOCK_GROUP_SIZE, BLOCK_HW_SIZE], dtype=tl.float32)
    dx_part3 = tl.zeros([BLOCK_GROUP_SIZE, BLOCK_HW_SIZE], dtype=tl.float32)
    for off in range(0, HW, BLOCK_HW_SIZE):
        hw_offset = off + tl.arange(0, BLOCK_HW_SIZE)
        hw_mask = hw_offset < HW
        xy_offset = pid * num_elements + group_offset[:, None] * HW + hw_offset[None, :]
        xy_mask = wb_mask[:, None] & hw_mask[None, :]

        dY_val = tl.load(grad_y + xy_offset, mask=xy_mask, other=0.0).to(tl.float32)
        X_val = tl.load(X + xy_offset, mask=xy_mask, other=0.0).to(tl.float32)

        x_hat = tl.where(xy_mask, rstd * (X_val - mean), 0.0)
        dx_hat = weight * dY_val
        dx_part2 += dx_hat
        dx_part3 += dx_hat * x_hat

    dx_2 = tl.sum(dx_part2)
    dx_3 = tl.sum(dx_part3)

    for off in range(0, HW, BLOCK_HW_SIZE):
        hw_offset = off + tl.arange(0, BLOCK_HW_SIZE)
        hw_mask = hw_offset < HW
        xy_offset = pid * num_elements + group_offset[:, None] * HW + hw_offset[None, :]
        xy_mask = wb_mask[:, None] & hw_mask[None, :]

        dY_val = tl.load(grad_y + xy_offset, mask=xy_mask, other=0.0).to(tl.float32)
        X_val = tl.load(X + xy_offset, mask=xy_mask, other=0.0).to(tl.float32)

        x_hat = tl.where(xy_mask, rstd * (X_val - mean), 0.0)
        dx_hat = weight * dY_val
        dx = rstd * (dx_hat - (dx_2 + x_hat * dx_3) / num_elements)
        grad_x_offset = tl.where(xy_mask, xy_offset, -1)

        tl.store(grad_x + grad_x_offset, dx, xy_mask)


@libentry()
@triton.jit
def weight_bias_backward_kernel(
    dY,
    X,
    Mean,
    Rstd,
    dW,
    dB,
    num_groups,
    group_size,
    N,
    C,
    HW,
    BLOCK_N: tl.constexpr,
    BLOCK_HW: tl.constexpr,
):
    pid = tle.program_id(0)
    group = pid // group_size
    n_offset = tl.arange(0, BLOCK_N)
    hw_offset = tl.arange(0, BLOCK_HW)
    xy_mask = n_offset[:, None] < N and hw_offset[None, :] < HW
    mr_mask = n_offset < N

    mean_ptr = Mean + group + n_offset * num_groups
    rstd_ptr = Rstd + group + n_offset * num_groups

    dY_ptr = dY + pid * HW + n_offset[:, None] * C * HW + hw_offset[None, :]
    x_ptr = X + pid * HW + n_offset[:, None] * C * HW + hw_offset[None, :]

    grad_y = tl.load(dY_ptr, mask=xy_mask, other=0.0).to(tl.float32)
    x = tl.load(x_ptr, mask=xy_mask, other=0.0)
    x_f32 = x.to(tl.float32)
    mean = tl.load(mean_ptr, mask=mr_mask, other=0.0).to(tl.float32)[:, None]
    rstd = tl.load(rstd_ptr, mask=mr_mask, other=0.0).to(tl.float32)[:, None]

    if dW is not None:
        dw = tl.sum((x_f32 - mean) * rstd * grad_y)
        tl.store(dW + pid, dw.to(x.dtype))
    if dB is not None:
        db = tl.sum(grad_y)
        tl.store(dB + pid, db.to(x.dtype))


@libentry()
@triton.jit
def weight_bias_backward_kernel_loop(
    dY,
    X,
    Mean,
    Rstd,
    dW,
    dB,
    num_groups,
    group_size,
    N,
    C,
    HW,
    BLOCK_N: tl.constexpr,
    BLOCK_HW: tl.constexpr,
):
    pid = tle.program_id(0)
    group = pid // group_size

    grad_y_tile = tl.zeros((BLOCK_N, BLOCK_HW), dtype=tl.float32)  # grad_y_tile
    dw_tile = tl.zeros((BLOCK_N, BLOCK_HW), dtype=tl.float32)
    for start_n in range(0, N, BLOCK_N):
        n_offset = start_n + tl.arange(0, BLOCK_N)

        mean_ptr = Mean + group + n_offset * num_groups
        rstd_ptr = Rstd + group + n_offset * num_groups
        mr_mask = n_offset < N
        mean = tl.load(mean_ptr, mask=mr_mask, other=0.0).to(tl.float32)[:, None]
        rstd = tl.load(rstd_ptr, mask=mr_mask, other=0.0).to(tl.float32)[:, None]

        for start_hw in range(0, HW, BLOCK_HW):
            hw_offset = start_hw + tl.arange(0, BLOCK_HW)
            xy_mask = n_offset[:, None] < N and hw_offset[None, :] < HW
            dY_ptr = dY + pid * HW + n_offset[:, None] * C * HW + hw_offset[None, :]
            grad_y = tl.load(dY_ptr, mask=xy_mask, other=0.0).to(tl.float32)
            grad_y_tile += grad_y

            x_ptr = X + pid * HW + n_offset[:, None] * C * HW + hw_offset[None, :]
            x = tl.load(x_ptr, mask=xy_mask, other=0.0)
            x_f32 = x.to(tl.float32)
            dw_tile += (x_f32 - mean) * rstd * grad_y

    dw = tl.sum(dw_tile)
    db = tl.sum(grad_y_tile)
    tl.store(dW + pid, dw)
    tl.store(dB + pid, db)


def group_norm(input, weight, bias, N, C, HxW, group, eps=1e-05):
    logger.debug("GEMS GROUPNORM FORWARD")

    group_size = triton.cdiv(C, group)
    input = input.contiguous()
    weight = None if weight is None else weight.contiguous()
    bias = None if bias is None else bias.contiguous()

    y = torch.empty_like(input)
    mean = torch.empty((N, group), dtype=input.dtype, device=input.device)
    rstd = torch.empty((N, group), dtype=input.dtype, device=input.device)

    grid = (N * group,)
    with torch_device_fn.device(input.device):
        if N == 1 and C == 64 and HxW == 1024 and group == 64:
            os.environ["TRITONXPU_OTHER_SIM"] = "1"
            os.environ["TRITONXPU_STORE_MASK_SIM"] = "1"
        group_norm_kernel[grid](
            input,
            y,
            weight,
            bias,
            mean,
            rstd,
            group_size,
            C,
            HxW,
            group,
            eps,
            BLOCK_GROUP_SIZE=triton.next_power_of_2(group_size),
            BLOCK_HW_SIZE=triton.next_power_of_2(HxW),
        )
        if "TRITONXPU_OTHER_SIM" in os.environ:
            del os.environ["TRITONXPU_OTHER_SIM"]
        if "TRITONXPU_STORE_MASK_SIM" in os.environ:
            del os.environ["TRITONXPU_STORE_MASK_SIM"]

    return y, mean, rstd


def group_norm_backward(
    grad_out, input, mean, rstd, weight, N, C, HxW, group, output_mask
):
    logger.debug("GEMS GROUPNORM BACKWARD")

    grad_out = grad_out.contiguous()
    input = input.contiguous()
    mean = mean.contiguous()
    rstd = rstd.contiguous()
    weight = None if weight is None else weight.contiguous()
    group_size = triton.cdiv(C, group)

    if output_mask[0]:
        grad_inp = torch.empty_like(input)
        grid = (N * group,)
        with torch_device_fn.device(input.device):
            import os

            os.environ["TRITONXPU_OTHER_SIM"] = "1"
            os.environ["TRITONXPU_STORE_MASK_SIM"] = "1"
            group_norm_backward_kernel[grid](
                grad_out,
                input,
                weight,
                mean,
                rstd,
                group,
                group_size,
                grad_inp,
                C,
                HxW,
                BLOCK_GROUP_SIZE=triton.next_power_of_2(group_size),
                BLOCK_HW_SIZE=triton.next_power_of_2(HxW),
                isCloseUnrollControl=True,
            )
            if "TRITONXPU_OTHER_SIM" in os.environ:
                del os.environ["TRITONXPU_OTHER_SIM"]
            if "TRITONXPU_STORE_MASK_SIM" in os.environ:
                del os.environ["TRITONXPU_STORE_MASK_SIM"]

    else:
        grad_inp = None

    if output_mask[1] is False and output_mask[2] is False:
        return grad_inp, None, None

    weight_grad = torch.empty_like(weight) if output_mask[1] else None
    bias_grad = torch.empty_like(weight) if output_mask[2] else None
    with torch_device_fn.device(input.device):
        if N == 32 and C == 32 and HxW == 1024 and group == 8:
            weight_bias_backward_kernel_loop[(C, 1, 1)](
                grad_out,
                input,
                mean,
                rstd,
                weight_grad,
                bias_grad,
                group,
                group_size,
                N,
                C,
                HxW,
                BLOCK_N=1,
                BLOCK_HW=triton.next_power_of_2(HxW),
                isCloseUnrollControl=True,
                isCloseCoreTiling=True,
            )
        else:
            if output_mask[1] is True and output_mask[2] is True:
                isCloseUnrollControl = True
            weight_bias_backward_kernel[(C, 1, 1)](
                grad_out,
                input,
                mean,
                rstd,
                weight_grad,
                bias_grad,
                group,
                group_size,
                N,
                C,
                HxW,
                BLOCK_N=triton.next_power_of_2(N),
                BLOCK_HW=triton.next_power_of_2(HxW),
                isCloseUnrollControl=isCloseUnrollControl,
            )
    return grad_inp, weight_grad, bias_grad



"""
Kernels for convolution.
"""


import triton
import triton.language as tl

from .utils import allow_tf32, get_n_stages


def conv2d_forward_config(
    BLOCK_SIZE_BATCH_HEIGHT_WIDTH: int,
    BLOCK_SIZE_IN_FEAT: int,
    BLOCK_SIZE_OUT_FEAT: int,
    n_warps: int = 4,
    n_stages: int = 2,
    ) -> triton.Config:
    """
    Creates a triton.Config object for conv2d_forward_kernel
    given meta-parameters for auto-tuning.

    Args:
        BLOCK_SIZE_BATCH_HEIGHT_WIDTH: Block size across the batch, height, and
            width dimensions.
        BLOCK_SIZE_IN_FEAT: Block size across the input feature dimension.
        BLOCK_SIZE_OUT_FEAT: Block size across the output feature dimension.
        n_warps: Number of warps to use for the kernel when compiled for GPUs.
        n_stages: Number of stages the compiler uses to software-pipeline.

    Returns:
        Kernel configuration.
    """
    return triton.Config({'BLOCK_SIZE_BATCH_HEIGHT_WIDTH': BLOCK_SIZE_BATCH_HEIGHT_WIDTH,
                          'BLOCK_SIZE_IN_FEAT': BLOCK_SIZE_IN_FEAT,
                          'BLOCK_SIZE_OUT_FEAT': BLOCK_SIZE_OUT_FEAT},
                          num_warps=n_warps, num_stages=get_n_stages(n_stages))


@triton.autotune(
    configs=[
        conv2d_forward_config(128, 32, 128, n_warps=8, n_stages=2),
        conv2d_forward_config(256, 32, 64, n_warps=8, n_stages=2),
        conv2d_forward_config(256, 32, 32, n_warps=4, n_stages=4),
        conv2d_forward_config(256, 64, 32, n_warps=4, n_stages=4),
        conv2d_forward_config(256, 32, 16, n_warps=2, n_stages=4),
        conv2d_forward_config(64, 32, 128, n_warps=8, n_stages=4),
        conv2d_forward_config(128, 32, 64, n_warps=4, n_stages=4),
        conv2d_forward_config(64, 32, 64, n_warps=4, n_stages=4),
        conv2d_forward_config(128, 32, 16, n_warps=4, n_stages=4),
        conv2d_forward_config(128, 128, 128, n_warps=8, n_stages=3),
        conv2d_forward_config(256, 128, 64, n_warps=8, n_stages=3),
        conv2d_forward_config(256, 128, 32, n_warps=4, n_stages=4),
        conv2d_forward_config(64, 128, 128, n_warps=4, n_stages=4),
        conv2d_forward_config(128, 128, 64, n_warps=4, n_stages=4),
        conv2d_forward_config(128, 64, 32, n_warps=2, n_stages=4),
        conv2d_forward_config(64, 64, 64, n_warps=2, n_stages=4),
    ],
    key=['batch_dim', 'in_feat_dim', 'in_height', 'in_width',
         'out_feat_dim', 'out_height', 'out_width',
         'kernel_height', 'kernel_width',
         'stride_height', 'stride_width',
         'padding_height', 'padding_width',
         'groups', 'fp16'],
)
@triton.heuristics({'tf32': lambda _: allow_tf32()})
@triton.jit
def conv2d_forward_kernel(
    input_pointer, weight_pointer, output_pointer,
    batch_dim, in_feat_dim, in_height, in_width,
    out_feat_dim, out_height, out_width,
    input_batch_stride, input_in_feat_stride, input_height_stride, input_width_stride,
    weight_out_feat_stride, weight_in_feat_stride, weight_height_stride, weight_width_stride,
    output_batch_stride, output_out_feat_stride, output_height_stride, output_width_stride,
    kernel_height: tl.constexpr, kernel_width: tl.constexpr,
    stride_height: tl.constexpr, stride_width: tl.constexpr,
    padding_height: tl.constexpr, padding_width: tl.constexpr,
    groups: tl.constexpr, fp16: tl.constexpr, tf32: tl.constexpr,
    BLOCK_SIZE_BATCH_HEIGHT_WIDTH: tl.constexpr, BLOCK_SIZE_IN_FEAT: tl.constexpr,
    BLOCK_SIZE_OUT_FEAT: tl.constexpr,
    ):
    """
    2D-convolves over the input using weights.

    Args:
        input_pointer: Pointer to the input to convolve over.
            The input must be of shape [batch_dim, in_feat_dim, in_height, in_width].
        weight_pointer: Pointer to the weights input is convolved over by.
            The weights must be of shape [out_feat_dim, in_feat_dim, kernel_height, kernel_width].
        output_pointer: Pointer to a container the result is written to.
            The container must be of shape [batch_dim, out_feat_dim, out_height, out_width].
        batch_dim: Batch dimension of the input and output.
        in_feat_dim: Dimensionality of the input features.
        in_height: Input height.
        in_width: Input width.
        out_feat_dim: Dimensionality of the output features.
        out_height: Output height.
        out_width: Output width.
        input_batch_stride: Stride necessary to jump one element along the
            input's batch dimension.
        input_in_feat_stride: Stride necessary to jump one element along the
            input's feature dimension.
        input_height_stride: Stride necessary to jump one element along the
            input's height dimension.
        input_width_stride: Stride necessary to jump one element along the
            input's width dimension.
        weight_out_feat_stride: Stride necessary to jump one element along the
            weights' output feature dimension.
        weight_in_feat_stride: Stride necessary to jump one element along the
            weights' input feature dimension.
        weight_height_stride: Stride necessary to jump one element along the
            weights' height dimension.
        weight_width_stride: Stride necessary to jump one element along the
            weights' width dimension.
        output_batch_stride: Stride necessary to jump one element along the
            output's batch dimension.
        output_out_feat_stride: Stride necessary to jump one element along the
            output's feature dimension.
        output_height_stride: Stride necessary to jump one element along the
            output's height dimension.
        output_width_stride: Stride necessary to jump one element along the
            output's width dimension.
        kernel_height: Kernel height.
        kernel_width: Kernel width.
        stride_height: Stride of kernel across the height dimension.
        stride_width: Stride of kernel across the width dimension.
        padding_height: Padding applied to the input across the height dimension.
        padding_width: Padding applied to the input across the width dimension.
        groups: Number of groups for the convolution.
        fp16: Flag for loading the input and weights in FP16.
        tf32: Flag for performing matrix products in TF32.
        BLOCK_SIZE_BATCH_HEIGHT_WIDTH: Block size across the batch, height, and
            width dimensions.
        BLOCK_SIZE_IN_FEAT: Block size across the input feature dimension.
        BLOCK_SIZE_OUT_FEAT: Block size across the output feature dimension.
    """
    batch_height_width_pid = tl.program_id(0)
    out_feat_pid = tl.program_id(1)
    group_pid = tl.program_id(2)

    in_group_dim = in_feat_dim // groups
    out_group_dim = out_feat_dim // groups

    batch_height_width_offset = (batch_height_width_pid * BLOCK_SIZE_BATCH_HEIGHT_WIDTH +
                                 tl.arange(0, BLOCK_SIZE_BATCH_HEIGHT_WIDTH))
    batch_height_offset = batch_height_width_offset // out_width
    batch_offset = batch_height_offset // out_height

    output_feat_offset = (out_feat_pid * BLOCK_SIZE_OUT_FEAT +
                          tl.arange(0, BLOCK_SIZE_OUT_FEAT))
    output_height_offset = batch_height_offset % out_height
    output_width_offset = batch_height_width_offset % out_width

    input_pointer += (input_batch_stride * batch_offset +
                      input_in_feat_stride * group_pid * in_group_dim)[:, None]
    weight_pointer += (weight_out_feat_stride * output_feat_offset +
                       weight_out_feat_stride * group_pid * out_group_dim)[None, :]

    accum = tl.zeros((BLOCK_SIZE_BATCH_HEIGHT_WIDTH, BLOCK_SIZE_OUT_FEAT),
                     dtype=tl.float32)

    for h in range(kernel_height):
        for w in range(kernel_width):
            for c in range(0, in_group_dim, BLOCK_SIZE_IN_FEAT):
                input_feat_offset = c + tl.arange(0, BLOCK_SIZE_IN_FEAT)
                input_height_offset = (h - padding_height +
                                       stride_height * output_height_offset)
                input_width_offset = (w - padding_width +
                                      stride_width * output_width_offset)

                curr_input_pointer = (input_pointer +
                                     (input_in_feat_stride * input_feat_offset)[None, :] +
                                     (input_height_stride * input_height_offset)[:, None] +
                                     (input_width_stride * input_width_offset)[:, None])
                curr_weight_pointer = (weight_pointer +
                                      (weight_in_feat_stride * input_feat_offset)[:, None] +
                                      (weight_height_stride * h) +
                                      (weight_width_stride * w))

                input_mask = ((batch_offset < batch_dim)[:, None] &
                              (input_feat_offset < in_group_dim)[None, :] &
                              (0 <= input_height_offset)[:, None] &
                              (input_height_offset < in_height)[:, None] &
                              (0 <= input_width_offset)[:, None] &
                              (input_width_offset < in_width)[:, None])
                weight_mask = ((input_feat_offset < in_group_dim)[:, None] &
                               (output_feat_offset < out_group_dim)[None, :])

                input_block = tl.load(curr_input_pointer, mask=input_mask)
                weight_block = tl.load(curr_weight_pointer, mask=weight_mask)

                if fp16:
                    input_block = input_block.to(tl.float16)
                    weight_block = weight_block.to(tl.float16)

                accum += tl.dot(input_block, weight_block, allow_tf32=tf32)

    output_pointer += ((output_batch_stride * batch_offset)[:, None] +
                       (output_out_feat_stride * (group_pid * out_group_dim + output_feat_offset))[None, :] +
                       (output_height_stride * output_height_offset)[:, None] +
                       (output_width_stride * output_width_offset)[:, None])
    output_mask = ((batch_offset < batch_dim)[:, None] &
                   (output_feat_offset < out_group_dim)[None, :] &
                   (output_height_offset < out_height)[:, None] &
                   (output_width_offset < out_width)[:, None])

    tl.store(output_pointer, accum, mask=output_mask)



"""
NeuroBrix Prism — Universal Hardware Auto-Detection (OS-First Architecture)

Detects CPU + GPU hardware and generates a hardware profile YAML.

Detection flow:
  1. Detect OS (Linux / macOS / Windows)
  2. Detect CPU (always — every machine has one)
  3. Detect GPUs via OS-specific cascade
  4. Detect interconnect (multi-GPU only)
  5. Build profile (GPU or CPU-only)

Supports:
  CPU        — x86_64 (AVX2/AVX512/AMX), ARM64 (NEON), RISC-V
  NVIDIA     — nvidia-smi
  AMD        — rocm-smi / rocminfo
  Intel      — xpu-smi / sycl-ls
  Apple      — system_profiler / MPS
  Tenstorrent — tt-smi
  Moore Threads — mtsmi
  Biren      — brsmi
  Iluvatar   — ixsmi
  Hygon DCU  — rocm-smi (DTK fork)
  Cambricon  — cnmon

Called when --hardware is omitted. Creates config/hardware/default.yml once.
"""

import platform
import re
import subprocess
import yaml
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple


# ============================================================================
# CONSTANTS
# ============================================================================

# Same directory as all other hardware profiles (config/hardware/)
HARDWARE_DIR = Path(__file__).parent.parent.parent / "config" / "hardware"
DEFAULT_PROFILE_PATH = HARDWARE_DIR / "default.yml"

# ---------------------------------------------------------------------------
# NVIDIA: compute capability → architecture name
# Source: CUDA Toolkit docs — static hardware fact, nvidia-smi reports CC not name
# ---------------------------------------------------------------------------
_NVIDIA_ARCH_MAP = {
    "3": "kepler",
    "5": "maxwell",
    "6": "pascal",
    "7": "volta",       # 7.0 = V100
    "8": "ampere",      # 8.0 = A100, 8.6 = RTX 3090
    "9": "hopper",      # 9.0 = H100
    "10": "blackwell",  # 10.0 = B200
}
_NVIDIA_ARCH_REFINE = {
    "7.5": "turing",    # RTX 2080, T4
    "8.9": "ada",       # RTX 4090, L40S
}

# NVIDIA: compute capability major → supported dtypes
_NVIDIA_DTYPE_MAP = {
    "3": ["float32"],
    "5": ["float32", "float16"],
    "6": ["float32", "float16"],
    "7": ["float32", "float16"],
    "8": ["float32", "float16", "bfloat16"],
    "9": ["float32", "float16", "bfloat16", "float8_e4m3fn"],
    "10": ["float32", "float16", "bfloat16", "float8_e4m3fn"],
}

# Architecture → preferred dtype (tensor core sweet spot)
_PREFERRED_DTYPE = {
    # NVIDIA
    "kepler": "float32", "maxwell": "float16", "pascal": "float16",
    "volta": "float16", "turing": "float16",
    "ampere": "bfloat16", "ada": "bfloat16",
    "hopper": "bfloat16", "blackwell": "bfloat16",
    # AMD
    "cdna": "float16", "cdna2": "float16", "cdna3": "bfloat16",
    "rdna2": "float16", "rdna3": "float16", "rdna4": "float16",
    # Intel
    "ponte_vecchio": "bfloat16", "flex": "float16",
    "alchemist": "float16", "battlemage": "float16", "xe": "float16",
    # Apple (M1=fp16, M2+=bf16 — refined below by chip generation)
    "apple_silicon": "float16",
    # Others
    "wormhole": "float16", "grayskull": "float16", "blackhole": "bfloat16",
    "musa": "float16", "birensupa": "float16",
    "corex": "float16", "hygon_dcu": "float16",
    "mlu": "float16",
}

# ---------------------------------------------------------------------------
# AMD: rocminfo gfx ID → architecture name
# ---------------------------------------------------------------------------
_AMD_GFX_MAP = {
    "gfx900": "vega10",     # MI25
    "gfx906": "vega20",     # MI50/MI60
    "gfx908": "cdna",       # MI100
    "gfx90a": "cdna2",      # MI210/MI250/MI250X
    "gfx940": "cdna3",      # MI300A
    "gfx941": "cdna3",      # MI300A variant
    "gfx942": "cdna3",      # MI300X
    "gfx1030": "rdna2",     # RX 6900 XT
    "gfx1100": "rdna3",     # RX 7900 XTX
    "gfx1101": "rdna3",     # RX 7900 XT
    "gfx1150": "rdna4",     # RX 9070 (projected)
    "gfx1200": "rdna4",     # RX 9070
}

# AMD: architecture → supported dtypes
_AMD_DTYPE_MAP = {
    "cdna3": ["float32", "float16", "bfloat16", "float8_e4m3fn"],
    "cdna2": ["float32", "float16", "bfloat16"],
    "cdna": ["float32", "float16"],
    "vega20": ["float32", "float16"],
    "vega10": ["float32", "float16"],
}

# NVLink bandwidth per generation (per link, bidirectional)
_NVLINK_BW_PER_LINK = {
    1: 40,    # Pascal P100
    2: 50,    # Volta V100
    3: 50,    # Ampere A100
    4: 100,   # Hopper H100
    5: 200,   # Blackwell B200
}

# ---------------------------------------------------------------------------
# PCI vendor IDs — for lspci fallback detection
# ---------------------------------------------------------------------------
_PCI_VENDOR_MAP = {
    "10de": "nvidia",
    "1002": "amd",
    "8086": "intel",
    "106b": "apple",
    "1e52": "moore_threads",
    "1e56": "tenstorrent",
    "1f36": "biren",
    "d100": "iluvatar",
    "cabc": "cambricon",
}

# PCIe link speed → version mapping
_PCIE_SPEED_MAP = {
    "2.5": "1.0", "5.0": "2.0", "8.0": "3.0",
    "16.0": "4.0", "32.0": "5.0", "64.0": "6.0",
}

# PCIe version → x16 bidirectional bandwidth (GB/s)
_PCIE_BW_MAP = {
    "1.0": 8, "2.0": 16, "3.0": 32,
    "4.0": 64, "5.0": 128, "6.0": 256,
}


# ============================================================================
# PUBLIC API
# ============================================================================

def get_or_create_default_profile() -> str:
    """
    Ensure default.yml exists in config/hardware/ (same dir as all profiles).

    Returns:
        The hardware_id string "default" (usable with load_profile("default")).
    """
    if DEFAULT_PROFILE_PATH.exists():
        return "default"

    print("   [Auto-detect] No --hardware specified, detecting system hardware...")
    profile_data = detect_hardware()

    HARDWARE_DIR.mkdir(parents=True, exist_ok=True)

    class _NoAliasDumper(yaml.SafeDumper):
        def ignore_aliases(self, data: object) -> bool:  # noqa: ARG002
            return True

    with open(DEFAULT_PROFILE_PATH, "w") as f:
        summary = profile_data.get("notes", "Auto-detected hardware profile")
        first_line = summary.strip().split("\n")[0]
        f.write(f"# Hardware Profile: {first_line}\n")
        f.write("# Auto-generated by NeuroBrix hardware detection\n")
        f.write("# Regenerate: delete this file and run neurobrix without --hardware\n\n")
        yaml.dump(profile_data, f, Dumper=_NoAliasDumper,
                  default_flow_style=False, sort_keys=False, allow_unicode=True)

    print(f"   [Auto-detect] Created profile: {DEFAULT_PROFILE_PATH}")
    return "default"


def load_default_profile():
    """
    Load the default hardware profile as a PrismProfile.

    Creates default.yml in config/hardware/ if it doesn't exist,
    then loads it through the standard load_profile() path.
    """
    get_or_create_default_profile()
    from neurobrix.core.prism.loader import load_profile
    return load_profile("default")


# ============================================================================
# MAIN DETECTION ORCHESTRATOR (OS-FIRST)
# ============================================================================

def detect_hardware() -> Dict[str, Any]:
    """
    Detect all hardware (CPU + GPUs + interconnects). Returns a hardware profile dict.

    OS-first detection:
    1. Detect OS → route to OS-specific detectors
    2. CPU detection (always — every machine has a CPU)
    3. GPU detection via OS-specific cascade
    4. Interconnect detection (multi-GPU only)
    """
    os_type = platform.system()  # "Linux", "Darwin", "Windows"

    # --- CPU (always) ---
    cpu = _detect_cpu(os_type)

    # --- GPUs (OS-specific cascade) ---
    devices, brand = _detect_gpus(os_type)

    # --- System vendor ---
    vendor = _detect_system_vendor(os_type)

    # --- Totals ---
    num_gpus = len(devices)
    total_vram_mb = sum(d["memory_mb"] for d in devices)
    total_vram_gb = round(total_vram_mb / 1024, 1) if total_vram_mb > 0 else 0
    ram_gb = round(cpu.get("ram_mb", 0) / 1024, 1) if cpu.get("ram_mb") else 0

    # --- Preferred dtype ---
    if devices:
        arch = devices[0].get("architecture", "unknown")
        preferred_dtype = _PREFERRED_DTYPE.get(arch, "float16")
        # Apple M2+: prefer bf16 (same exponent range as fp32, no overflow)
        if arch == "apple_silicon" and "bfloat16" in devices[0].get("supports_dtypes", []):
            preferred_dtype = "bfloat16"
    else:
        # CPU-only: derive from CPU features
        preferred_dtype = _cpu_preferred_dtype(cpu)

    # --- Profile ID ---
    if devices:
        model_short = _shorten_gpu_name(devices[0].get("model", "gpu"))
        if num_gpus == 1:
            mem_gb = round(devices[0]["memory_mb"] / 1024)
            profile_id = f"auto-{model_short}-{mem_gb}g"
        else:
            profile_id = f"auto-{num_gpus}x{model_short}-{total_vram_gb}g"
    else:
        cpu_short = _shorten_cpu_name(cpu.get("model", "cpu"))
        profile_id = f"auto-{cpu_short}-cpu-{int(ram_gb)}g"

    # --- Interconnect ---
    if num_gpus > 1:
        interconnect, pcie_fallback = _detect_interconnect(devices, brand)
    elif num_gpus == 1:
        pcie_ver = devices[0].get("pcie_version", "3.0")
        pcie_bw = _PCIE_BW_MAP.get(pcie_ver, 32)
        interconnect = {"groups": []}
        pcie_fallback = {"version": pcie_ver, "lanes": 16, "bandwidth_gbps": pcie_bw}
    else:
        # CPU-only: no PCIe relevant
        interconnect = {"groups": []}
        pcie_fallback = {"version": "0.0", "lanes": 0, "bandwidth_gbps": 0}

    # --- Topology label ---
    if num_gpus == 0:
        topology_label = "CPU-Only"
    elif num_gpus == 1:
        topology_label = "Single-GPU"
    else:
        topology_label = "Multi-GPU"

    # --- Summary ---
    summary: Dict[str, Any] = {
        "total_gpus": num_gpus,
        "total_vram_gb": total_vram_gb,
        "total_ram_gb": ram_gb,
        "topology": topology_label,
    }

    # --- Notes ---
    notes = _build_notes(cpu, devices, brand, vendor, interconnect)

    profile: Dict[str, Any] = {
        "id": profile_id,
        "vendor": vendor,
        "preferred_dtype": preferred_dtype,
        "summary": summary,
        "cpu": cpu,
        "devices": devices,
        "interconnect": interconnect,
        "pcie_fallback": pcie_fallback,
        "notes": notes,
    }

    return profile


# ============================================================================
# CPU DETECTION (per-OS)
# ============================================================================

def _detect_cpu(os_type: str) -> Dict[str, Any]:
    """Detect CPU info. Always succeeds — every machine has a CPU."""
    if os_type == "Linux":
        return _detect_cpu_linux()
    elif os_type == "Darwin":
        return _detect_cpu_darwin()
    elif os_type == "Windows":
        return _detect_cpu_windows()
    # Unknown OS — minimal info
    return {
        "model": platform.processor() or "Unknown CPU",
        "cores": 1,
        "threads": 1,
        "ram_mb": 4096,
        "architecture": platform.machine(),
        "features": [],
    }


def _detect_cpu_linux() -> Dict[str, Any]:
    """Detect CPU on Linux via /proc/cpuinfo + /proc/meminfo + lscpu."""
    model = "Unknown CPU"
    features: List[str] = []
    cores = 1
    threads = 1
    ram_mb = 4096
    arch = platform.machine()  # x86_64, aarch64, riscv64

    # --- /proc/cpuinfo ---
    try:
        cpuinfo = Path("/proc/cpuinfo").read_text()
        # Model name (x86)
        m = re.search(r'^model name\s*:\s*(.+)$', cpuinfo, re.MULTILINE)
        if m:
            model = m.group(1).strip()

        # ARM: look for "CPU implementer" + "Hardware" or "model name"
        if model == "Unknown CPU":
            m = re.search(r'^Hardware\s*:\s*(.+)$', cpuinfo, re.MULTILINE)
            if m:
                model = m.group(1).strip()

        # CPU flags (x86) or Features (ARM)
        flags_match = re.search(r'^(?:flags|Features)\s*:\s*(.+)$', cpuinfo, re.MULTILINE)
        if flags_match:
            all_flags = flags_match.group(1).strip().split()
            # Extract relevant features for inference
            relevant = {
                "avx", "avx2", "avx512f", "avx512_vnni", "avx512_bf16",
                "amx_bf16", "amx_int8", "amx_tile",
                "fma", "sse4_1", "sse4_2", "f16c",
                "asimd", "fphp", "asimdhp", "sve", "sve2",  # ARM
            }
            features = [f for f in all_flags if f in relevant]
    except (FileNotFoundError, PermissionError):
        pass

    # --- /proc/meminfo ---
    try:
        meminfo = Path("/proc/meminfo").read_text()
        m = re.search(r'^MemTotal:\s*(\d+)\s*kB', meminfo, re.MULTILINE)
        if m:
            ram_mb = int(m.group(1)) // 1024
    except (FileNotFoundError, PermissionError):
        pass

    # --- lscpu for core/thread counts ---
    try:
        result = subprocess.run(
            ["lscpu"], capture_output=True, text=True, timeout=5,
        )
        if result.returncode == 0:
            for line in result.stdout.split("\n"):
                if "Core(s) per socket:" in line:
                    val = line.split(":")[-1].strip()
                    try:
                        sockets_cores = int(val)
                    except ValueError:
                        sockets_cores = 1
                elif "Socket(s):" in line:
                    val = line.split(":")[-1].strip()
                    try:
                        sockets = int(val)
                    except ValueError:
                        sockets = 1
                elif "CPU(s):" in line and "NUMA" not in line and "On-line" not in line:
                    val = line.split(":")[-1].strip()
                    try:
                        threads = int(val)
                    except ValueError:
                        pass

            # cores = cores_per_socket * sockets
            try:
                cores = sockets_cores * sockets  # type: ignore[possibly-undefined]
            except NameError:
                cores = threads
    except (FileNotFoundError, subprocess.TimeoutExpired):
        # Fallback: nproc
        try:
            result = subprocess.run(
                ["nproc"], capture_output=True, text=True, timeout=5,
            )
            if result.returncode == 0:
                threads = int(result.stdout.strip())
                cores = threads  # Can't distinguish without lscpu
        except Exception:
            pass

    return {
        "model": model,
        "cores": cores,
        "threads": threads,
        "ram_mb": ram_mb,
        "architecture": arch,
        "features": features,
    }


def _detect_cpu_darwin() -> Dict[str, Any]:
    """Detect CPU on macOS via sysctl."""
    model = "Unknown CPU"
    cores = 1
    threads = 1
    ram_mb = 8192
    arch = platform.machine()  # arm64 or x86_64
    features: List[str] = []

    # Model name
    model = _sysctl_str("machdep.cpu.brand_string") or "Unknown CPU"

    # If Apple Silicon and brand_string is empty, try hw.model
    if model == "Unknown CPU":
        model = _sysctl_str("hw.model") or "Unknown CPU"

    # Core counts
    cores = _sysctl_int("hw.physicalcpu") or 1
    threads = _sysctl_int("hw.logicalcpu") or cores

    # RAM
    memsize = _sysctl_int("hw.memsize")
    if memsize:
        ram_mb = memsize // (1024 * 1024)

    # CPU features
    if arch == "arm64":
        # Apple Silicon always has NEON + FP16
        features = ["neon", "fp16"]
    else:
        # Intel Mac — check for AVX
        leaf7 = _sysctl_str("machdep.cpu.leaf7_features") or ""
        ext_features = _sysctl_str("machdep.cpu.extfeatures") or ""
        cpu_features = _sysctl_str("machdep.cpu.features") or ""
        all_feat = (cpu_features + " " + leaf7 + " " + ext_features).lower()
        relevant = {"avx", "avx2", "avx512f", "fma", "sse4.1", "sse4.2", "f16c"}
        features = [f for f in relevant if f in all_feat]

    return {
        "model": model,
        "cores": cores,
        "threads": threads,
        "ram_mb": ram_mb,
        "architecture": arch,
        "features": features,
    }


def _detect_cpu_windows() -> Dict[str, Any]:
    """Detect CPU on Windows via wmic or PowerShell."""
    model = "Unknown CPU"
    cores = 1
    threads = 1
    ram_mb = 8192
    arch = platform.machine()
    features: List[str] = []

    # --- wmic CPU ---
    try:
        result = subprocess.run(
            ["wmic", "cpu", "get",
             "Name,NumberOfCores,NumberOfLogicalProcessors",
             "/format:csv"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode == 0:
            for line in result.stdout.strip().split("\n"):
                parts = [p.strip() for p in line.split(",")]
                if len(parts) >= 4 and parts[1]:
                    model = parts[1]
                    try:
                        cores = int(parts[2])
                    except (ValueError, IndexError):
                        pass
                    try:
                        threads = int(parts[3])
                    except (ValueError, IndexError):
                        pass
    except (FileNotFoundError, subprocess.TimeoutExpired):
        # PowerShell fallback
        try:
            result = subprocess.run(
                ["powershell", "-Command",
                 "Get-CimInstance Win32_Processor | "
                 "Select-Object Name,NumberOfCores,NumberOfLogicalProcessors | "
                 "ConvertTo-Json"],
                capture_output=True, text=True, timeout=10,
            )
            if result.returncode == 0:
                import json
                data = json.loads(result.stdout)
                if isinstance(data, list):
                    data = data[0]
                model = data.get("Name", model)
                cores = data.get("NumberOfCores", cores)
                threads = data.get("NumberOfLogicalProcessors", threads)
        except Exception:
            pass

    # --- wmic RAM ---
    try:
        result = subprocess.run(
            ["wmic", "OS", "get", "TotalVisibleMemorySize", "/format:csv"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode == 0:
            for line in result.stdout.strip().split("\n"):
                parts = [p.strip() for p in line.split(",")]
                if len(parts) >= 2:
                    try:
                        ram_kb = int(parts[-1])
                        ram_mb = ram_kb // 1024
                    except ValueError:
                        pass
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass

    # Infer features from model name (best-effort on Windows)
    model_lower = model.lower()
    if "13th gen" in model_lower or "14th gen" in model_lower or "ultra" in model_lower:
        features = ["avx2", "avx512f"]
    elif "12th gen" in model_lower or "11th gen" in model_lower:
        features = ["avx2", "avx512f"]
    elif any(gen in model_lower for gen in ["5th gen", "sapphire", "emerald", "granite"]):
        features = ["avx2", "avx512f", "amx_bf16", "amx_int8"]
    elif "ryzen" in model_lower or "epyc" in model_lower:
        features = ["avx2"]
    elif "apple" in model_lower or "m1" in model_lower or "m2" in model_lower:
        features = ["neon", "fp16"]

    return {
        "model": model,
        "cores": cores,
        "threads": threads,
        "ram_mb": ram_mb,
        "architecture": arch,
        "features": features,
    }


def _cpu_preferred_dtype(cpu: Dict[str, Any]) -> str:
    """Determine preferred dtype for CPU-only inference."""
    features = cpu.get("features", [])
    # Intel AMX — native bf16 matrix ops
    if "amx_bf16" in features or "amx_tile" in features:
        return "bfloat16"
    # AVX512 BF16 extension
    if "avx512_bf16" in features:
        return "bfloat16"
    # ARM FP16 arithmetic (Apple Silicon, Graviton3+)
    if "fp16" in features or "fphp" in features or "asimdhp" in features:
        return "float16"
    return "float32"


# ============================================================================
# GPU DETECTION (OS-FIRST DISPATCH)
# ============================================================================

def _detect_gpus(os_type: str) -> Tuple[List[Dict[str, Any]], str]:
    """Detect GPUs using OS-specific cascade. Returns (devices, brand)."""
    if os_type == "Linux":
        return _detect_gpus_linux()
    elif os_type == "Darwin":
        return _detect_gpus_darwin()
    elif os_type == "Windows":
        return _detect_gpus_windows()
    return [], "unknown"


def _detect_gpus_linux() -> Tuple[List[Dict[str, Any]], str]:
    """
    Linux GPU detection cascade:
    nvidia-smi → rocm → xpu-smi → tt-smi → chinese SMI → lspci → PyTorch
    """
    for detector, brand in [
        (_parse_nvidia_smi, "nvidia"),
        (_parse_rocm, "amd"),
        (_parse_xpu, "intel"),
        (_parse_tenstorrent, "tenstorrent"),
        (_parse_moore_threads, "moore_threads"),
        (_parse_biren, "biren"),
        (_parse_iluvatar, "iluvatar"),
        (_parse_cambricon, "cambricon"),
    ]:
        devices = detector()
        if devices:
            return devices, brand

    # lspci fallback (Linux only)
    devices, brand = _parse_lspci()
    if devices:
        return devices, brand

    # PyTorch backend fallback
    return _detect_torch_fallback()


def _detect_gpus_darwin() -> Tuple[List[Dict[str, Any]], str]:
    """
    macOS GPU detection:
    system_profiler → PyTorch MPS → PyTorch CUDA (Hackintosh edge case)
    """
    devices = _parse_system_profiler()
    if devices:
        brand = devices[0].get("brand", "apple")
        return devices, brand

    # PyTorch MPS fallback
    return _detect_torch_fallback()


def _detect_gpus_windows() -> Tuple[List[Dict[str, Any]], str]:
    """
    Windows GPU detection cascade:
    nvidia-smi → rocm-smi → xpu-smi → WMI/PowerShell
    """
    for detector, brand in [
        (_parse_nvidia_smi, "nvidia"),
        (_parse_rocm, "amd"),
        (_parse_xpu, "intel"),
    ]:
        devices = detector()
        if devices:
            return devices, brand

    # WMI fallback (Windows only)
    devices, brand = _parse_wmi()
    if devices:
        return devices, brand

    # PyTorch backend fallback
    return _detect_torch_fallback()


# ============================================================================
# GPU VENDOR PARSERS (shared by Linux + Windows where CLI is identical)
# ============================================================================

def _parse_nvidia_smi() -> List[Dict[str, Any]]:
    """Detect NVIDIA GPUs via nvidia-smi."""
    try:
        result = subprocess.run(
            ["nvidia-smi",
             "--query-gpu=index,name,memory.total,compute_cap,pci.bus_id",
             "--format=csv,noheader,nounits"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            return []
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return []

    devices = []
    for line in result.stdout.strip().split("\n"):
        if not line.strip():
            continue
        parts = [p.strip() for p in line.split(",")]
        if len(parts) < 5:
            continue

        idx = int(parts[0])
        name = parts[1]
        mem_mb = int(float(parts[2]))
        cc = parts[3]
        pci_bus = parts[4]

        arch = _nvidia_cc_to_arch(cc)
        cc_major = cc.split(".")[0]
        dtypes = _NVIDIA_DTYPE_MAP.get(cc_major, ["float32", "float16"])
        pcie_ver = _detect_pcie_version_sysfs(pci_bus)

        devices.append({
            "index": idx,
            "brand": "nvidia",
            "model": name,
            "memory_mb": mem_mb,
            "compute_capability": cc,
            "supports_dtypes": dtypes,
            "architecture": arch,
            "pcie_version": pcie_ver,
        })

    return devices


def _parse_rocm() -> List[Dict[str, Any]]:
    """Detect AMD GPUs via rocminfo + rocm-smi."""
    devices = _parse_rocminfo()
    if devices:
        return devices
    return _parse_rocmsmi()


def _parse_rocminfo() -> List[Dict[str, Any]]:
    """Detect AMD GPUs via rocminfo (best for architecture detection)."""
    try:
        result = subprocess.run(
            ["rocminfo"], capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            return []
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return []

    devices = []
    current: Dict[str, str] = {}
    idx = 0

    for line in result.stdout.split("\n"):
        line = line.strip()

        if line.startswith("Name:") and "gfx" in line:
            gfx_id = line.split(":")[-1].strip()
            current["gfx_id"] = gfx_id

        elif line.startswith("Marketing Name:") and current:
            current["name"] = line.split(":", 1)[-1].strip()

        elif line.startswith("Device Type:") and "GPU" in line:
            current["is_gpu"] = "true"

        elif line.startswith("Size:") and current.get("is_gpu"):
            size_str = line.split(":")[-1].strip().split("(")[0].strip()
            try:
                mem_bytes = int(size_str)
                if mem_bytes > 1_000_000_000:
                    current["memory_bytes"] = str(mem_bytes)
            except ValueError:
                pass

        elif line == "" and current.get("is_gpu") and current.get("gfx_id"):
            gfx_id = current["gfx_id"]
            name = current.get("name", f"AMD GPU ({gfx_id})")
            mem_bytes = int(current.get("memory_bytes", "0"))
            mem_mb = mem_bytes // (1024 * 1024) if mem_bytes > 0 else 0

            if "hygon" in name.lower():
                current = {}
                continue

            arch = _AMD_GFX_MAP.get(gfx_id, _amd_gfx_to_arch(gfx_id))
            dtypes = _AMD_DTYPE_MAP.get(arch, ["float32", "float16"])

            devices.append({
                "index": idx,
                "brand": "amd",
                "model": name,
                "memory_mb": mem_mb if mem_mb > 0 else 16384,
                "compute_capability": gfx_id,
                "supports_dtypes": dtypes,
                "architecture": arch,
                "pcie_version": "4.0",
            })
            idx += 1
            current = {}

    return devices


def _parse_rocmsmi() -> List[Dict[str, Any]]:
    """Detect AMD GPUs via rocm-smi --showproductname."""
    try:
        result = subprocess.run(
            ["rocm-smi", "--showproductname"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            return []
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return []

    devices = []
    idx = 0
    for line in result.stdout.split("\n"):
        match = re.search(r'GPU\[(\d+)\].*?:\s*(.+)', line)
        if not match:
            continue
        gpu_idx = int(match.group(1))
        name = match.group(2).strip()
        mem_mb = _amd_get_vram_mb(gpu_idx)
        arch = _amd_name_to_arch(name)
        dtypes = _AMD_DTYPE_MAP.get(arch, ["float32", "float16"])

        devices.append({
            "index": gpu_idx,
            "brand": "amd",
            "model": name,
            "memory_mb": mem_mb,
            "compute_capability": "0.0",
            "supports_dtypes": dtypes,
            "architecture": arch,
            "pcie_version": "4.0",
        })
        idx += 1

    return devices


def _parse_xpu() -> List[Dict[str, Any]]:
    """Detect Intel GPUs via xpu-smi or sycl-ls."""
    devices = _parse_xpusmi()
    if devices:
        return devices
    return _parse_sycl()


def _parse_xpusmi() -> List[Dict[str, Any]]:
    """Detect Intel GPUs via xpu-smi discovery."""
    try:
        result = subprocess.run(
            ["xpu-smi", "discovery", "--json"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            result = subprocess.run(
                ["xpu-smi", "discovery"],
                capture_output=True, text=True, timeout=10,
            )
            if result.returncode != 0:
                return []
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return []

    devices = []
    idx = 0

    # Try JSON
    try:
        import json
        data = json.loads(result.stdout)
        for dev in data if isinstance(data, list) else []:
            name = dev.get("device_name", "Intel GPU")
            mem_mb = dev.get("memory_physical_size_byte", 0) // (1024 * 1024)
            pci_addr = dev.get("pci_bdf_address", "")
            pcie_ver = _detect_pcie_version_sysfs(pci_addr) if pci_addr else "4.0"
            arch = _intel_name_to_arch(name)

            devices.append({
                "index": idx,
                "brand": "intel",
                "model": name,
                "memory_mb": mem_mb if mem_mb > 0 else 16384,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16"],
                "architecture": arch,
                "pcie_version": pcie_ver,
            })
            idx += 1
        if devices:
            return devices
    except Exception:
        pass

    # Text parse fallback
    for line in result.stdout.split("\n"):
        line = line.strip()
        if "Device Name" in line and ":" in line:
            name = line.split(":", 1)[-1].strip()
            arch = _intel_name_to_arch(name)
            devices.append({
                "index": idx,
                "brand": "intel",
                "model": name,
                "memory_mb": 16384,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16"],
                "architecture": arch,
                "pcie_version": "4.0",
            })
            idx += 1

    return devices


def _parse_sycl() -> List[Dict[str, Any]]:
    """Detect Intel GPUs via sycl-ls."""
    try:
        result = subprocess.run(
            ["sycl-ls"], capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            return []
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return []

    devices = []
    idx = 0
    seen: set[str] = set()
    for line in result.stdout.split("\n"):
        if "gpu" in line.lower() and ("intel" in line.lower() or "arc" in line.lower()):
            name = "Intel GPU"
            match = re.search(r'\]\s+(.+)', line)
            if match:
                name = match.group(1).strip()
                if name in seen:
                    continue
                seen.add(name)

            arch = _intel_name_to_arch(name)
            devices.append({
                "index": idx,
                "brand": "intel",
                "model": name,
                "memory_mb": 16384,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16"],
                "architecture": arch,
                "pcie_version": "4.0",
            })
            idx += 1

    return devices


def _parse_tenstorrent() -> List[Dict[str, Any]]:
    """Detect Tenstorrent accelerators via tt-smi."""
    try:
        result = subprocess.run(
            ["tt-smi", "--json"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            result = subprocess.run(
                ["tt-smi"], capture_output=True, text=True, timeout=10,
            )
            if result.returncode != 0:
                return []
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return []

    devices = []
    idx = 0

    # Try JSON
    try:
        import json
        data = json.loads(result.stdout)
        for dev in data.get("devices", data if isinstance(data, list) else []):
            name = dev.get("board_type", dev.get("name", "Tenstorrent Accelerator"))
            arch = "wormhole" if "wormhole" in name.lower() else \
                   "blackhole" if "blackhole" in name.lower() else "grayskull"
            mem_mb = 12288 if "n150" in name.lower() else \
                     24576 if "n300" in name.lower() else 12288

            devices.append({
                "index": idx,
                "brand": "tenstorrent",
                "model": name,
                "memory_mb": mem_mb,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16", "bfloat16"],
                "architecture": arch,
                "pcie_version": "4.0",
            })
            idx += 1
        if devices:
            return devices
    except Exception:
        pass

    # Text parse fallback
    for line in result.stdout.split("\n"):
        line_lower = line.lower()
        if "wormhole" in line_lower or "grayskull" in line_lower or "blackhole" in line_lower:
            name = line.strip().split("|")[1].strip() if "|" in line else line.strip()
            arch = "wormhole" if "wormhole" in line_lower else \
                   "blackhole" if "blackhole" in line_lower else "grayskull"
            mem_mb = 12288 if "n150" in line_lower else \
                     24576 if "n300" in line_lower else 12288

            devices.append({
                "index": idx,
                "brand": "tenstorrent",
                "model": name,
                "memory_mb": mem_mb,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16", "bfloat16"],
                "architecture": arch,
                "pcie_version": "4.0",
            })
            idx += 1

    return devices


def _parse_moore_threads() -> List[Dict[str, Any]]:
    """Detect Moore Threads GPUs via mtsmi."""
    return _run_generic_smi(["mtsmi", "-q"], ["mtsmi"], "moore_threads", "musa")


def _parse_biren() -> List[Dict[str, Any]]:
    """Detect Biren GPUs via brsmi."""
    return _run_generic_smi(["brsmi", "-q"], ["brsmi"], "biren", "birensupa")


def _parse_iluvatar() -> List[Dict[str, Any]]:
    """Detect Iluvatar CoreX GPUs via ixsmi."""
    return _run_generic_smi(["ixsmi", "-q"], ["ixsmi"], "iluvatar", "corex")


def _parse_cambricon() -> List[Dict[str, Any]]:
    """Detect Cambricon MLU via cnmon."""
    return _run_generic_smi(["cnmon", "info"], ["cnmon"], "cambricon", "mlu")


def _run_generic_smi(
    primary_cmd: List[str],
    fallback_cmd: List[str],
    brand: str,
    default_arch: str,
) -> List[Dict[str, Any]]:
    """Run a vendor SMI tool and parse its output."""
    try:
        result = subprocess.run(
            primary_cmd, capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            result = subprocess.run(
                fallback_cmd, capture_output=True, text=True, timeout=10,
            )
            if result.returncode != 0:
                return []
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return []

    return _parse_smi_output(result.stdout, brand, default_arch)


# ============================================================================
# MACOS-ONLY: system_profiler
# ============================================================================

def _parse_system_profiler() -> List[Dict[str, Any]]:
    """Detect GPUs on macOS via system_profiler SPDisplaysDataType."""
    try:
        result = subprocess.run(
            ["system_profiler", "SPDisplaysDataType", "-json"],
            capture_output=True, text=True, timeout=15,
        )
        if result.returncode != 0:
            return []
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return []

    try:
        import json
        data = json.loads(result.stdout)
    except Exception:
        return []

    displays = data.get("SPDisplaysDataType", [])
    devices = []
    idx = 0

    for gpu in displays:
        name = gpu.get("sppci_model", gpu.get("_name", "Unknown GPU"))
        vendor_str = gpu.get("spdisplays_vendor", "").lower()

        # Determine brand
        if "apple" in vendor_str or "apple" in name.lower():
            brand = "apple"
            arch = "apple_silicon"
            # Apple Silicon: unified memory — get from sysctl
            mem_mb = _detect_apple_memory_mb()
            dtypes = ["float32", "float16"]
            if _apple_supports_bf16(name):
                dtypes.append("bfloat16")
            pcie_ver = "N/A"
        elif "amd" in vendor_str or "ati" in vendor_str:
            brand = "amd"
            arch = _amd_name_to_arch(name)
            # VRAM from system_profiler
            vram_str = gpu.get("spdisplays_vram", gpu.get("spdisplays_vram_shared", "0"))
            mem_mb = _parse_mac_vram(vram_str)
            dtypes = ["float32", "float16"]
            pcie_ver = "3.0"
        elif "intel" in vendor_str:
            brand = "intel"
            arch = "xe"
            vram_str = gpu.get("spdisplays_vram_shared", gpu.get("spdisplays_vram", "0"))
            mem_mb = _parse_mac_vram(vram_str)
            dtypes = ["float32"]
            pcie_ver = "3.0"
        else:
            brand = "apple"
            arch = "apple_silicon"
            mem_mb = _detect_apple_memory_mb()
            dtypes = ["float32", "float16"]
            pcie_ver = "N/A"

        # Check Metal support
        metal_support = gpu.get("spdisplays_metal", "")
        if "supported" in str(metal_support).lower():
            if "float16" not in dtypes:
                dtypes.append("float16")

        devices.append({
            "index": idx,
            "brand": brand,
            "model": name,
            "memory_mb": mem_mb if mem_mb > 0 else 8192,
            "compute_capability": "0.0",
            "supports_dtypes": dtypes,
            "architecture": arch,
            "pcie_version": pcie_ver,
        })
        idx += 1

    return devices


# ============================================================================
# WINDOWS-ONLY: WMI / PowerShell
# ============================================================================

def _parse_wmi() -> Tuple[List[Dict[str, Any]], str]:
    """Detect GPUs on Windows via WMI (Win32_VideoController)."""
    try:
        result = subprocess.run(
            ["wmic", "path", "Win32_VideoController", "get",
             "Name,AdapterRAM,PNPDeviceID", "/format:csv"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            return _parse_powershell_gpu()
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return _parse_powershell_gpu()

    devices = []
    brand = "unknown"
    idx = 0

    for line in result.stdout.strip().split("\n"):
        parts = [p.strip() for p in line.split(",")]
        if len(parts) < 4:
            continue
        # CSV: Node,AdapterRAM,Name,PNPDeviceID
        adapter_ram = parts[1]
        name = parts[2]
        pnp_id = parts[3] if len(parts) > 3 else ""

        if not name or name == "Name":
            continue

        # Skip Microsoft Basic Display Adapter (software renderer)
        if "microsoft" in name.lower() and "basic" in name.lower():
            continue

        # Detect brand from PNPDeviceID (contains VEN_XXXX)
        ven_match = re.search(r'VEN_([0-9A-Fa-f]{4})', pnp_id)
        if ven_match:
            ven_id = ven_match.group(1).lower()
            detected_brand = _PCI_VENDOR_MAP.get(ven_id, "unknown")
            if detected_brand != "unknown":
                brand = detected_brand

        # Memory
        try:
            mem_bytes = int(adapter_ram)
            mem_mb = mem_bytes // (1024 * 1024) if mem_bytes > 0 else 0
        except (ValueError, TypeError):
            mem_mb = 0

        # WMI AdapterRAM is capped at 4GB (32-bit field) — unreliable for modern GPUs
        if mem_mb > 0 and mem_mb <= 4096 and "rtx" in name.lower():
            mem_mb = 0  # Will use fallback

        devices.append({
            "index": idx,
            "brand": brand if brand != "unknown" else _infer_brand_from_name(name),
            "model": name,
            "memory_mb": mem_mb if mem_mb > 0 else 8192,
            "compute_capability": "0.0",
            "supports_dtypes": ["float32", "float16"],
            "architecture": "unknown",
            "pcie_version": "3.0",
        })
        idx += 1

    return devices, brand


def _parse_powershell_gpu() -> Tuple[List[Dict[str, Any]], str]:
    """Detect GPUs on Windows via PowerShell Get-CimInstance."""
    try:
        result = subprocess.run(
            ["powershell", "-Command",
             "Get-CimInstance Win32_VideoController | "
             "Select-Object Name,AdapterRAM,PNPDeviceID | ConvertTo-Json"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            return [], "unknown"
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return [], "unknown"

    try:
        import json
        data = json.loads(result.stdout)
        if not isinstance(data, list):
            data = [data]
    except Exception:
        return [], "unknown"

    devices = []
    brand = "unknown"
    idx = 0

    for gpu in data:
        name = gpu.get("Name", "Unknown GPU")
        if "microsoft" in name.lower() and "basic" in name.lower():
            continue

        pnp_id = gpu.get("PNPDeviceID", "")
        ven_match = re.search(r'VEN_([0-9A-Fa-f]{4})', pnp_id)
        if ven_match:
            ven_id = ven_match.group(1).lower()
            detected = _PCI_VENDOR_MAP.get(ven_id, "unknown")
            if detected != "unknown":
                brand = detected

        adapter_ram = gpu.get("AdapterRAM", 0)
        mem_mb = int(adapter_ram) // (1024 * 1024) if adapter_ram else 0

        devices.append({
            "index": idx,
            "brand": brand if brand != "unknown" else _infer_brand_from_name(name),
            "model": name,
            "memory_mb": mem_mb if mem_mb > 0 else 8192,
            "compute_capability": "0.0",
            "supports_dtypes": ["float32", "float16"],
            "architecture": "unknown",
            "pcie_version": "3.0",
        })
        idx += 1

    return devices, brand


# ============================================================================
# LINUX-ONLY: lspci FALLBACK
# ============================================================================

def _parse_lspci() -> Tuple[List[Dict[str, Any]], str]:
    """Detect GPUs via lspci PCI vendor IDs + sysfs (Linux only)."""
    try:
        result = subprocess.run(
            ["lspci", "-Dnn"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            return [], "unknown"
    except (FileNotFoundError, subprocess.TimeoutExpired):
        return [], "unknown"

    devices = []
    brand = "unknown"
    idx = 0

    for line in result.stdout.split("\n"):
        if not any(cls in line for cls in ["VGA", "3D controller", "Processing accelerator"]):
            continue

        vid_match = re.search(r'\[([0-9a-f]{4}):([0-9a-f]{4})\]', line)
        if not vid_match:
            continue
        vendor_id = vid_match.group(1)
        detected_brand = _PCI_VENDOR_MAP.get(vendor_id)
        if not detected_brand:
            continue

        if devices and brand != detected_brand:
            continue

        brand = detected_brand

        pci_addr_match = re.match(r'([0-9a-f:.]+)', line)
        pci_addr = pci_addr_match.group(1) if pci_addr_match else ""

        name_match = re.search(r'\]:\s*(.+?)\s*\[', line)
        name = name_match.group(1).strip() if name_match else f"{brand} GPU"

        pcie_ver = _detect_pcie_version_sysfs(pci_addr) if pci_addr else "4.0"

        mem_mb = 0
        if brand == "amd":
            sysfs_mem = Path(f"/sys/bus/pci/devices/{pci_addr}/mem_info_vram_total")
            if sysfs_mem.exists():
                try:
                    mem_mb = int(sysfs_mem.read_text().strip()) // (1024 * 1024)
                except (ValueError, PermissionError):
                    pass

        devices.append({
            "index": idx,
            "brand": brand,
            "model": name,
            "memory_mb": mem_mb if mem_mb > 0 else 16384,
            "compute_capability": "0.0",
            "supports_dtypes": ["float32", "float16"],
            "architecture": "unknown",
            "pcie_version": pcie_ver,
        })
        idx += 1

    return devices, brand


# ============================================================================
# GENERIC SMI PARSER (shared by Moore Threads, Biren, Iluvatar, Cambricon)
# ============================================================================

def _parse_smi_output(
    output: str,
    brand: str,
    default_arch: str,
) -> List[Dict[str, Any]]:
    """
    Parse nvidia-smi-style output from vendor SMI tools.
    Most Chinese GPU vendors model their CLI after nvidia-smi.
    """
    devices = []
    idx = 0
    current_name = ""
    current_mem_mb = 0

    for line in output.split("\n"):
        line_stripped = line.strip()

        name_match = re.search(
            r'(?:Product Name|GPU Name|Device Name|Card Name)\s*:\s*(.+)',
            line_stripped, re.IGNORECASE,
        )
        if name_match:
            current_name = name_match.group(1).strip()

        mem_match = re.search(
            r'(?:Total|FB Memory|Memory Total)\s*:\s*(\d+)\s*(?:MiB|MB)',
            line_stripped, re.IGNORECASE,
        )
        if mem_match:
            current_mem_mb = int(mem_match.group(1))

        mem_bytes_match = re.search(
            r'(?:Total|Memory)\s*:\s*(\d+)\s*(?:bytes|B)\b',
            line_stripped, re.IGNORECASE,
        )
        if mem_bytes_match:
            val = int(mem_bytes_match.group(1))
            if val > 1_000_000_000:
                current_mem_mb = val // (1024 * 1024)

        mem_gb_match = re.search(
            r'(?:Total|Memory|VRAM)\s*:\s*(\d+)\s*(?:GiB|GB)',
            line_stripped, re.IGNORECASE,
        )
        if mem_gb_match:
            current_mem_mb = int(mem_gb_match.group(1)) * 1024

        table_match = re.match(r'\|\s*(\d+)\s*\|\s*(.+?)\s*\|', line_stripped)
        if table_match:
            row_name = table_match.group(2).strip()
            if row_name and not row_name.startswith("-") and not row_name.startswith("="):
                if not current_name:
                    current_name = row_name

        gpu_boundary = re.search(r'GPU\s*\[?(\d+)', line_stripped, re.IGNORECASE)
        if gpu_boundary and current_name and idx > 0:
            devices.append({
                "index": idx - 1,
                "brand": brand,
                "model": current_name,
                "memory_mb": current_mem_mb if current_mem_mb > 0 else 16384,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16"],
                "architecture": default_arch,
                "pcie_version": "4.0",
            })
            current_name = ""
            current_mem_mb = 0

        if gpu_boundary:
            idx = int(gpu_boundary.group(1)) + 1

    if current_name:
        devices.append({
            "index": max(0, idx - 1) if idx > 0 else 0,
            "brand": brand,
            "model": current_name,
            "memory_mb": current_mem_mb if current_mem_mb > 0 else 16384,
            "compute_capability": "0.0",
            "supports_dtypes": ["float32", "float16"],
            "architecture": default_arch,
            "pcie_version": "4.0",
        })

    return devices


# ============================================================================
# PYTORCH BACKEND FALLBACK (last resort — catches any GPU PyTorch can see)
# ============================================================================

def _detect_torch_fallback() -> Tuple[List[Dict[str, Any]], str]:
    """Last-resort detection via PyTorch backends."""
    try:
        import torch
    except ImportError:
        return [], "unknown"

    # CUDA (NVIDIA, AMD ROCm, Hygon DTK, Iluvatar IXUCA)
    if torch.cuda.is_available():
        devices = []
        is_hip = hasattr(torch, 'version') and hasattr(torch.version, 'hip') and torch.version.hip is not None  # type: ignore[union-attr]
        brand = "amd" if is_hip else "nvidia"

        for i in range(torch.cuda.device_count()):
            props = torch.cuda.get_device_properties(i)
            name = props.name
            mem_mb = props.total_memory // (1024 * 1024)
            cc = f"{props.major}.{props.minor}"
            name_lower = name.lower()

            if "hygon" in name_lower:
                brand = "amd"
                arch = "hygon_dcu"
                dtypes = ["float32", "float16"]
                cc = getattr(props, 'gcnArchName', "0.0")
            elif "iluvatar" in name_lower or "bi-v" in name_lower:
                brand = "iluvatar"
                arch = "corex"
                dtypes = ["float32", "float16"]
                cc = "0.0"
            elif is_hip:
                arch = _amd_name_to_arch(name)
                dtypes = _AMD_DTYPE_MAP.get(arch, ["float32", "float16"])
                cc = getattr(props, 'gcnArchName', cc)
            else:
                arch = _nvidia_cc_to_arch(cc)
                cc_major = str(props.major)
                dtypes = _NVIDIA_DTYPE_MAP.get(cc_major, ["float32", "float16"])

            devices.append({
                "index": i,
                "brand": brand,
                "model": name,
                "memory_mb": mem_mb,
                "compute_capability": cc,
                "supports_dtypes": dtypes,
                "architecture": arch,
                "pcie_version": "3.0",
            })
        return devices, brand

    # Intel XPU
    if hasattr(torch, 'xpu') and torch.xpu.is_available():  # type: ignore[attr-defined]
        devices = []
        for i in range(torch.xpu.device_count()):  # type: ignore[attr-defined]
            props = torch.xpu.get_device_properties(i)  # type: ignore[attr-defined]
            name = props.name
            mem_mb = props.total_memory // (1024 * 1024)
            arch = _intel_name_to_arch(name)

            devices.append({
                "index": i,
                "brand": "intel",
                "model": name,
                "memory_mb": mem_mb,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16"],
                "architecture": arch,
                "pcie_version": "4.0",
            })
        return devices, "intel"

    # Moore Threads (torch.musa)
    if hasattr(torch, 'musa') and torch.musa.is_available():  # type: ignore[attr-defined]
        devices = []
        for i in range(torch.musa.device_count()):  # type: ignore[attr-defined]
            props = torch.musa.get_device_properties(i)  # type: ignore[attr-defined]
            name = props.name
            mem_mb = props.total_memory // (1024 * 1024)
            devices.append({
                "index": i,
                "brand": "moore_threads",
                "model": name,
                "memory_mb": mem_mb,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16"],
                "architecture": "musa",
                "pcie_version": "4.0",
            })
        return devices, "moore_threads"

    # Cambricon MLU (torch.mlu)
    if hasattr(torch, 'mlu') and torch.mlu.is_available():  # type: ignore[attr-defined]
        devices = []
        for i in range(torch.mlu.device_count()):  # type: ignore[attr-defined]
            name = torch.mlu.get_device_name(i)  # type: ignore[attr-defined]
            props = torch.mlu.get_device_properties(i)  # type: ignore[attr-defined]
            mem_mb = props.total_memory // (1024 * 1024)
            devices.append({
                "index": i,
                "brand": "cambricon",
                "model": name,
                "memory_mb": mem_mb,
                "compute_capability": "0.0",
                "supports_dtypes": ["float32", "float16"],
                "architecture": "mlu",
                "pcie_version": "4.0",
            })
        return devices, "cambricon"

    # Apple MPS
    if hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
        chip_name = _detect_apple_chip()
        memory_mb = _detect_apple_memory_mb()

        # bf16 support: M2+ chips with macOS 14+ (Sonoma)
        supports_dtypes = ["float32", "float16"]
        if _apple_supports_bf16(chip_name):
            supports_dtypes.append("bfloat16")

        devices = [{
            "index": 0,
            "brand": "apple",
            "model": chip_name,
            "memory_mb": memory_mb,
            "compute_capability": "N/A",
            "supports_dtypes": supports_dtypes,
            "architecture": "apple_silicon",
            "pcie_version": "N/A",
            "unified_memory": True,
        }]
        return devices, "apple"

    return [], "unknown"


# ============================================================================
# INTERCONNECT DETECTION
# ============================================================================

def _detect_interconnect(
    devices: List[Dict[str, Any]],
    brand: str,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Detect GPU interconnect topology. Returns (interconnect, pcie_fallback)."""
    num_gpus = len(devices)
    pcie_ver = devices[0].get("pcie_version", "3.0") if devices else "3.0"
    pcie_bw = _PCIE_BW_MAP.get(pcie_ver, 32)

    pcie_fallback = {
        "version": pcie_ver,
        "lanes": 16,
        "bandwidth_gbps": pcie_bw,
    }

    if num_gpus <= 1:
        return {"groups": []}, pcie_fallback

    members = list(range(num_gpus))

    if brand == "nvidia":
        return _detect_nvidia_interconnect(devices, members, pcie_fallback)
    elif brand == "amd":
        return _detect_amd_interconnect(members, pcie_fallback)
    elif brand == "tenstorrent":
        return _detect_tenstorrent_interconnect(members, pcie_fallback)
    elif brand == "cambricon":
        return _detect_cambricon_interconnect(members, pcie_fallback)

    return {
        "groups": [{
            "index": 0,
            "members": members,
            "type": "cpu_bounce",
            "tech": "pcie",
            "bandwidth_gbps": pcie_bw,
        }],
    }, pcie_fallback


def _detect_nvidia_interconnect(
    devices: List[Dict[str, Any]],
    members: List[int],
    pcie_fallback: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Detect NVIDIA NVLink topology via nvidia-smi topo -m."""
    num_gpus = len(devices)

    try:
        result = subprocess.run(
            ["nvidia-smi", "topo", "-m"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode != 0:
            raise RuntimeError("nvidia-smi topo failed")
    except Exception:
        return _pcie_interconnect(members, pcie_fallback), pcie_fallback

    nvlink_pairs: List[Tuple[int, int, int]] = []
    has_nvlink = False

    for line in result.stdout.strip().split("\n"):
        line = line.strip()
        if not line.startswith("GPU"):
            continue
        parts = line.split()
        if len(parts) < 2:
            continue
        match = re.match(r'GPU(\d+)', parts[0])
        if not match:
            continue
        gpu_a = int(match.group(1))

        for col_idx, val in enumerate(parts[1:], start=0):
            if val.startswith("NV"):
                has_nvlink = True
                num_links = int(val[2:]) if len(val) > 2 and val[2:].isdigit() else 1
                gpu_b = col_idx
                if gpu_a < gpu_b:
                    nvlink_pairs.append((gpu_a, gpu_b, num_links))

    if not has_nvlink:
        return _pcie_interconnect(members, pcie_fallback), pcie_fallback

    cc = devices[0].get("compute_capability", "7.0")
    cc_major = int(cc.split(".")[0])
    nvlink_gen = max(1, cc_major - 5)
    bw_per_link = _NVLINK_BW_PER_LINK.get(nvlink_gen, 50)

    # Aggregate NVLink bandwidth per GPU
    links_per_gpu: Dict[int, int] = {}
    for a, b, n in nvlink_pairs:
        links_per_gpu[a] = links_per_gpu.get(a, 0) + n
        links_per_gpu[b] = links_per_gpu.get(b, 0) + n
    max_links = max(links_per_gpu.values()) if links_per_gpu else 1
    total_bw = max_links * bw_per_link

    all_connected = all(
        any(a == i and b == j for a, b, _ in nvlink_pairs)
        for i in range(num_gpus) for j in range(i + 1, num_gpus)
    )

    topo_type = "full_mesh" if all_connected and num_gpus > 1 else \
                "nvlink_p2p" if num_gpus == 2 else "full_mesh"

    return {
        "groups": [{
            "index": 0,
            "members": members,
            "type": topo_type,
            "tech": "nvlink",
            "bandwidth_gbps": total_bw,
        }],
    }, pcie_fallback


def _detect_amd_interconnect(
    members: List[int],
    pcie_fallback: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Detect AMD Infinity Fabric / xGMI interconnect."""
    try:
        result = subprocess.run(
            ["rocm-smi", "--showtopo"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode == 0 and "XGMI" in result.stdout.upper():
            return {
                "groups": [{
                    "index": 0,
                    "members": members,
                    "type": "full_mesh",
                    "tech": "xgmi",
                    "bandwidth_gbps": 200,
                }],
            }, pcie_fallback
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass

    for card in sorted(Path("/sys/class/drm/").glob("card*")):
        hive_path = card / "device" / "xgmi_hive_id"
        if hive_path.exists():
            try:
                hive_id = hive_path.read_text().strip()
                if hive_id and hive_id != "0":
                    return {
                        "groups": [{
                            "index": 0,
                            "members": members,
                            "type": "full_mesh",
                            "tech": "xgmi",
                            "bandwidth_gbps": 200,
                        }],
                    }, pcie_fallback
            except (PermissionError, ValueError):
                pass

    return _pcie_interconnect(members, pcie_fallback), pcie_fallback


def _detect_tenstorrent_interconnect(
    members: List[int],
    pcie_fallback: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Detect Tenstorrent ethernet mesh interconnect."""
    try:
        result = subprocess.run(
            ["tt-topology"], capture_output=True, text=True, timeout=10,
        )
        if result.returncode == 0 and "eth" in result.stdout.lower():
            return {
                "groups": [{
                    "index": 0,
                    "members": members,
                    "type": "full_mesh",
                    "tech": "ethernet_mesh",
                    "bandwidth_gbps": 100,
                }],
            }, pcie_fallback
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass
    return _pcie_interconnect(members, pcie_fallback), pcie_fallback


def _detect_cambricon_interconnect(
    members: List[int],
    pcie_fallback: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Detect Cambricon MLULink interconnect."""
    try:
        result = subprocess.run(
            ["cnmon", "mlulink", "-t"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode == 0 and "mlulink" in result.stdout.lower():
            return {
                "groups": [{
                    "index": 0,
                    "members": members,
                    "type": "full_mesh",
                    "tech": "mlulink",
                    "bandwidth_gbps": 200,
                }],
            }, pcie_fallback
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass
    return _pcie_interconnect(members, pcie_fallback), pcie_fallback


def _pcie_interconnect(
    members: List[int],
    pcie_fallback: Dict[str, Any],
) -> Dict[str, Any]:
    """Default PCIe-only interconnect group."""
    return {
        "groups": [{
            "index": 0,
            "members": members,
            "type": "cpu_bounce",
            "tech": "pcie",
            "bandwidth_gbps": pcie_fallback.get("bandwidth_gbps", 32),
        }],
    }


# ============================================================================
# SYSTEM VENDOR DETECTION (per-OS)
# ============================================================================

def _detect_system_vendor(os_type: str) -> str:
    """Detect system/motherboard vendor."""
    if os_type == "Darwin":
        return "apple"

    if os_type == "Linux":
        return _detect_vendor_linux()

    if os_type == "Windows":
        return _detect_vendor_windows()

    return "custom"


def _detect_vendor_linux() -> str:
    """Detect system vendor on Linux from DMI data."""
    paths = [
        "/sys/class/dmi/id/sys_vendor",
        "/sys/class/dmi/id/board_vendor",
        "/sys/class/dmi/id/chassis_vendor",
    ]
    known_vendors = {
        "dell": "dell", "hp": "hp", "hewlett": "hp",
        "lenovo": "lenovo", "supermicro": "supermicro",
        "asus": "asus", "gigabyte": "gigabyte", "msi": "msi",
        "apple": "apple", "inspur": "inspur", "huawei": "huawei",
        "sugon": "sugon", "h3c": "h3c",
    }
    for path in paths:
        try:
            with open(path) as f:
                raw = f.read().strip().lower()
            for keyword, name in known_vendors.items():
                if keyword in raw:
                    return name
            if raw:
                return raw
        except (FileNotFoundError, PermissionError):
            continue
    return "custom"


def _detect_vendor_windows() -> str:
    """Detect system vendor on Windows via wmic."""
    try:
        result = subprocess.run(
            ["wmic", "computersystem", "get", "Manufacturer", "/format:csv"],
            capture_output=True, text=True, timeout=10,
        )
        if result.returncode == 0:
            for line in result.stdout.strip().split("\n"):
                parts = [p.strip() for p in line.split(",")]
                if len(parts) >= 2 and parts[-1] and parts[-1] != "Manufacturer":
                    raw = parts[-1].lower()
                    known = {
                        "dell": "dell", "hp": "hp", "hewlett": "hp",
                        "lenovo": "lenovo", "asus": "asus",
                        "gigabyte": "gigabyte", "msi": "msi",
                        "apple": "apple", "microsoft": "microsoft",
                    }
                    for keyword, name in known.items():
                        if keyword in raw:
                            return name
                    return raw
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass
    return "custom"


# ============================================================================
# UNIVERSAL HELPERS
# ============================================================================

def _nvidia_cc_to_arch(cc: str) -> str:
    """Convert NVIDIA compute capability to architecture name."""
    if cc in _NVIDIA_ARCH_REFINE:
        return _NVIDIA_ARCH_REFINE[cc]
    major = cc.split(".")[0]
    return _NVIDIA_ARCH_MAP.get(major, "unknown")


def _amd_gfx_to_arch(gfx_id: str) -> str:
    """Map unknown gfx IDs to architecture by prefix pattern."""
    if gfx_id.startswith("gfx9"):
        major = int(gfx_id[3:5]) if len(gfx_id) >= 5 else 90
        if major >= 94:
            return "cdna3"
        elif major >= 90:
            return "cdna2" if "a" in gfx_id else "cdna"
        return "vega20"
    elif gfx_id.startswith("gfx12"):
        return "rdna4"
    elif gfx_id.startswith("gfx11"):
        return "rdna3"
    elif gfx_id.startswith("gfx10"):
        return "rdna2"
    return "rdna"


def _amd_name_to_arch(name: str) -> str:
    """Infer AMD architecture from GPU marketing name."""
    n = name.lower()
    if "mi300" in n:
        return "cdna3"
    elif "mi250" in n or "mi210" in n:
        return "cdna2"
    elif "mi100" in n:
        return "cdna"
    elif any(x in n for x in ["7900", "7800", "7700", "7600"]):
        return "rdna3"
    elif any(x in n for x in ["9070", "9060"]):
        return "rdna4"
    elif any(x in n for x in ["6900", "6800", "6700"]):
        return "rdna2"
    return "rdna"


def _amd_get_vram_mb(gpu_idx: int) -> int:
    """Get AMD GPU VRAM via sysfs or rocm-smi."""
    for card in sorted(Path("/sys/class/drm/").glob("card*")):
        vram_path = card / "device" / "mem_info_vram_total"
        if vram_path.exists():
            try:
                vram_bytes = int(vram_path.read_text().strip())
                if vram_bytes > 1_000_000_000:
                    return vram_bytes // (1024 * 1024)
            except (ValueError, PermissionError):
                pass

    try:
        result = subprocess.run(
            ["rocm-smi", "-d", str(gpu_idx), "--showmeminfo", "vram"],
            capture_output=True, text=True, timeout=5,
        )
        if result.returncode == 0:
            for line in result.stdout.split("\n"):
                if "Total" in line:
                    nums = re.findall(r'(\d+)', line)
                    if nums:
                        val = int(nums[-1])
                        if val > 1_000_000:
                            return val // (1024 * 1024)
    except Exception:
        pass

    return 16384


def _intel_name_to_arch(name: str) -> str:
    """Infer Intel GPU architecture from name."""
    n = name.lower()
    if any(x in n for x in ["b580", "b570", "battlemage"]):
        return "battlemage"
    elif any(x in n for x in ["a770", "a750", "a580", "a380", "alchemist"]):
        return "alchemist"
    elif "max" in n or "ponte" in n or "pvc" in n:
        return "ponte_vecchio"
    elif "flex" in n:
        return "flex"
    return "xe"


def _detect_pcie_version_sysfs(pci_addr: str) -> str:
    """Detect PCIe version from Linux sysfs for any vendor."""
    if not pci_addr or pci_addr == "N/A":
        return "3.0"
    try:
        addr = pci_addr.lower().strip()
        sysfs_path = Path(f"/sys/bus/pci/devices/{addr}/current_link_speed")
        if sysfs_path.exists():
            speed = sysfs_path.read_text().strip()
            gt_match = re.search(r'([\d.]+)\s*GT/s', speed)
            if gt_match:
                gt_val = gt_match.group(1)
                return _PCIE_SPEED_MAP.get(gt_val, "3.0")
    except Exception:
        pass
    return "3.0"


def _apple_supports_bf16(chip_name: str) -> bool:
    """
    Check if Apple Silicon chip supports bfloat16.

    bf16 requires M2 or later AND macOS 14+ (Sonoma).
    M1 family does NOT support bf16 in hardware.
    """
    import platform
    gen = _apple_chip_generation(chip_name)
    if gen < 2:
        return False

    # Check macOS version (need 14+ for bf16)
    try:
        mac_ver = platform.mac_ver()[0]  # e.g., "14.2.1"
        major = int(mac_ver.split('.')[0])
        return major >= 14
    except (ValueError, IndexError):
        return False


def _apple_chip_generation(chip_name: str) -> int:
    """
    Extract Apple Silicon generation number from chip name.

    "Apple M1 Max" → 1, "Apple M4 Pro" → 4, "Apple M5" → 5
    Returns 0 if not recognized.
    """
    import re
    m = re.search(r'\bm(\d+)', chip_name.lower())
    return int(m.group(1)) if m else 0


# Apple Silicon GPU specifications — from Apple official specs.
# Key: chip variant → (gpu_cores, memory_bandwidth_gbps, max_memory_gb)
# Sources: apple.com/newsroom, en.wikipedia.org/wiki/Apple_silicon
_APPLE_SILICON_SPECS = {
    # M1 family (2020-2021)
    "m1":           (8,   68.25, 16),
    "m1 pro":       (16,  200,   32),
    "m1 max":       (32,  400,   64),
    "m1 ultra":     (64,  800,   128),
    # M2 family (2022-2023)
    "m2":           (10,  100,   24),
    "m2 pro":       (19,  200,   32),
    "m2 max":       (38,  400,   96),
    "m2 ultra":     (76,  800,   192),
    # M3 family (2023-2024)
    "m3":           (10,  100,   24),
    "m3 pro":       (18,  150,   36),
    "m3 max":       (40,  400,   128),
    "m3 ultra":     (80,  800,   512),
    # M4 family (2024-2025)
    "m4":           (10,  120,   32),
    "m4 pro":       (20,  273,   64),
    "m4 max":       (40,  546,   128),
    "m4 ultra":     (80,  819,   256),
    # M5 family (2025-2026)
    "m5":           (10,  154,   32),
    "m5 pro":       (20,  307,   64),
    "m5 max":       (40,  614,   128),
    "m5 ultra":     (80,  921,   256),
}


def _lookup_apple_gpu_cores(chip_name: str) -> int:
    """Get GPU core count from chip name. Returns 8 as safe default."""
    key = _normalize_apple_chip_name(chip_name)
    if key in _APPLE_SILICON_SPECS:
        return _APPLE_SILICON_SPECS[key][0]
    return 8


def _lookup_apple_bandwidth(chip_name: str) -> float:
    """Get memory bandwidth in GB/s from chip name. Returns 100.0 as default."""
    key = _normalize_apple_chip_name(chip_name)
    if key in _APPLE_SILICON_SPECS:
        return _APPLE_SILICON_SPECS[key][1]
    return 100.0


def _normalize_apple_chip_name(chip_name: str) -> str:
    """
    Normalize chip name for lookup.

    "Apple M4 Max 40-core GPU" → "m4 max"
    "Apple M2" → "m2"
    """
    import re
    name = chip_name.lower()
    m = re.search(r'm(\d+)\s*(ultra|max|pro)?', name)
    if not m:
        return ""
    gen = m.group(1)
    variant = m.group(2) or ""
    return f"m{gen} {variant}".strip()


def _detect_apple_chip() -> str:
    """Detect Apple Silicon chip model."""
    return _sysctl_str("machdep.cpu.brand_string") or \
           _sysctl_str("hw.model") or "Apple Silicon"


def _detect_apple_memory_mb() -> int:
    """Detect Apple Silicon unified memory (shared CPU+GPU)."""
    memsize = _sysctl_int("hw.memsize")
    if memsize:
        return memsize // (1024 * 1024)
    return 8192


def _parse_mac_vram(vram_str: str) -> int:
    """Parse macOS VRAM string like '4 GB' or '1536 MB'."""
    if not vram_str:
        return 0
    m = re.search(r'(\d+)\s*(GB|MB)', str(vram_str), re.IGNORECASE)
    if m:
        val = int(m.group(1))
        unit = m.group(2).upper()
        return val * 1024 if unit == "GB" else val
    return 0


def _infer_brand_from_name(name: str) -> str:
    """Infer GPU brand from device name string."""
    n = name.lower()
    if any(x in n for x in ["nvidia", "geforce", "rtx", "gtx", "quadro", "tesla"]):
        return "nvidia"
    elif any(x in n for x in ["radeon", "amd", "instinct"]):
        return "amd"
    elif any(x in n for x in ["intel", "arc", "iris"]):
        return "intel"
    return "unknown"


def _shorten_gpu_name(name: str) -> str:
    """Shorten GPU name for profile ID."""
    name = name.lower()
    for prefix in ["nvidia ", "geforce ", "tesla ", "quadro ", "amd ", "intel ",
                    "radeon ", "instinct ", "arc ", "cambricon ", "moore threads ",
                    "iluvatar ", "biren ", "tenstorrent "]:
        name = name.replace(prefix, "")
    for suffix in ["-sxm2", "-sxm4", "-sxm5", "-sxm", "-pcie", "-hbm2",
                    "-hbm2e", "-hbm3", " ti", " super"]:
        name = name.replace(suffix, "")
    name = re.sub(r'[^a-z0-9]+', '-', name).strip('-')
    return name


def _shorten_cpu_name(name: str) -> str:
    """Shorten CPU name for profile ID."""
    name = name.lower()
    for prefix in ["intel(r) ", "amd ", "apple ", "core(tm) ", "xeon(r) ",
                    "processor ", "cpu "]:
        name = name.replace(prefix, "")
    # Remove frequency
    name = re.sub(r'@\s*[\d.]+\s*[gm]hz', '', name)
    name = re.sub(r'[^a-z0-9]+', '-', name).strip('-')
    # Truncate to reasonable length
    if len(name) > 20:
        name = name[:20].rstrip('-')
    return name


def _sysctl_str(key: str) -> Optional[str]:
    """Read a macOS sysctl string value."""
    try:
        result = subprocess.run(
            ["sysctl", "-n", key],
            capture_output=True, text=True, timeout=5,
        )
        if result.returncode == 0 and result.stdout.strip():
            return result.stdout.strip()
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass
    return None


def _sysctl_int(key: str) -> Optional[int]:
    """Read a macOS sysctl integer value."""
    val = _sysctl_str(key)
    if val:
        try:
            return int(val)
        except ValueError:
            pass
    return None


def _build_notes(
    cpu: Dict[str, Any],
    devices: List[Dict[str, Any]],
    brand: str,
    vendor: str,
    interconnect: Dict[str, Any],
) -> str:
    """Build human-readable notes string."""
    lines = []

    num_gpus = len(devices)

    if num_gpus == 0:
        lines.append(f"CPU-only system: {cpu.get('model', 'Unknown CPU')}")
        ram_gb = round(cpu.get('ram_mb', 0) / 1024, 1)
        lines.append(f"RAM: {ram_gb} GB, {cpu.get('cores', 1)} cores / {cpu.get('threads', 1)} threads")
        features = cpu.get("features", [])
        if features:
            lines.append(f"CPU features: {', '.join(features)}")
        lines.append("No GPU detected. Models run on CPU.")
    else:
        if num_gpus == 1:
            lines.append(f"1 x {devices[0].get('model', 'GPU')}")
        else:
            models: Dict[str, int] = {}
            for d in devices:
                m = d.get("model", "GPU")
                models[m] = models.get(m, 0) + 1
            parts = [f"{count} x {model}" for model, count in models.items()]
            lines.append(" + ".join(parts))

        lines.append(f"GPU brand: {brand}")
        arch = devices[0].get("architecture", "unknown")
        lines.append(f"Architecture: {arch}")

        # CPU info for offload context
        ram_gb = round(cpu.get('ram_mb', 0) / 1024, 1)
        lines.append(f"CPU: {cpu.get('model', 'Unknown')} ({cpu.get('cores', '?')} cores, {ram_gb} GB RAM)")

        groups = interconnect.get("groups", [])
        if groups:
            g = groups[0]
            tech = g.get("tech", "none")
            if tech not in ("none", "pcie"):
                bw = g.get("bandwidth_gbps", 0)
                lines.append(f"Interconnect: {tech} ({bw} GB/s)")

    if vendor != "custom":
        lines.append(f"System vendor: {vendor}")

    lines.append("Auto-generated by NeuroBrix hardware detection.")
    return "\n".join(lines)



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.utils import libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@libentry()
@triton.jit
def arange_func(
    y_ptr,
    start,
    end,
    step,
    size,
    BLOCK_SIZE: tl.constexpr,
    buffer_size_limit: tl.constexpr,
):
    pid = tle.program_id(0)
    y_ptr += pid * BLOCK_SIZE
    step_offset = pid * BLOCK_SIZE * step

    cols = tl.arange(0, BLOCK_SIZE)
    arange_val = cols * step + step_offset + start
    mask = cols + pid * BLOCK_SIZE
    tl.store(y_ptr + cols, arange_val, mask=mask < size)


def arange_start(
    start, end, step=1, *, dtype=None, layout=None, device=None, pin_memory=None
):
    logger.debug("GEMS ARANGE")
    if dtype is torch.int64:
        sgn = (step > 0) - (step < 0)
        size = (end - start + step - sgn) // step
    else:
        size = math.ceil((end - start) / step)

    cluster_num = 12
    tmp = torch.tensor([], dtype=dtype)
    BLOCK_SIZE = min(
        triton.next_power_of_2(triton.cdiv(size, cluster_num)),
        int(2048 * 64 / tmp.element_size()),
    )
    grid = triton.cdiv(size, BLOCK_SIZE)

    if dtype is None:
        dtype = torch.int64

    if pin_memory is None:
        pin_memory = False

    if device is None:
        device = (
            runtime.device.name
        )  # Note(Zhengzekang): Torch default value is CPU, but triton is target to GPU.

    result = torch.empty((size,), device=device, dtype=dtype, pin_memory=pin_memory)
    arange_func[grid,](
        result, start, end, step, size, BLOCK_SIZE, buffer_size_limit=2048
    )
    return result


def arange(end, *, dtype=None, layout=None, device=None, pin_memory=None):
    return arange_start(
        0, end, 1, dtype=dtype, layout=layout, device=device, pin_memory=pin_memory
    )
